/*
 * Copyright (c) 1993-2003
 *      Texas A&M University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Texas A&M University
 *      and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Developers:
 *     Russell Neeper, David K. Hess, Douglas Lee Schales, David R. Safford,
 *     Wm. Daryl Hawkins
 *
 * $Id: show.c,v 1.1.1.1 2004/01/06 21:09:54 daryl Exp $
 *
 */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <netdb.h>
#include <ctype.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <netgraph.h>

#include "/usr/src/sys/netgraph/ng_drawbridge.h"
#include "dbmgr.h"
#include "tokens.h"
#include "misc.h"

#define __SHOW__
#include "text.h"

static int lookupAccRej(int, char *);
static int queryAcceptTable(void);
static int queryRejectTable(void);
static int lookupOverride(char *);
static char print_line(PortListEntry *, u_short);
static int queryTCPresetTable(void);
static int queryClassTable(u_short, char *);
static int queryNetworkTable(char *, char, char *);
static void showStats(void);

/*------------------------------------------------------------*/

void
usageShow(int verbose)
{
	fprintf(theOutput, "%s%s",
		(verbose) ? showText[0] : showText[0]+1,
		(verbose) ? showText[1] : "");
}

/*------------------------------------------------------------*/

static int
lookupAccRej(int id, char *hoststr)
{
	struct in_addr hostaddr;
	rejaccTableReq req;
	u_int32_t host;
	u_char buffer[2*sizeof(struct ng_mesg) + sizeof(rejaccTableReq)];
	struct ng_mesg *const m = (struct ng_mesg *) buffer;
	char path[NG_PATHLEN+1];

	if (isalpha(hoststr[0])) {
		struct hostent *h;

		if ((h = gethostbyname(hoststr))) {
			host = ntohl(*((long *) h->h_addr_list[0]));
		} else {
			fprintf(theErrors, "   Unknown host '%s'\n", hoststr);
			return ERROR;
		}
	} else
		host = inet_network(hoststr);

	req.ptr = NULL;
	req.addr = host;
	hostaddr.s_addr = htonl(host); 
	if (id == T_REJTBL) {
		/* Send a message to drawbridge: to get stats */
		if (NgSendMsg(cskt,"drawbridge:",NGM_DRAWBRIDGE_COOKIE,
		    NGM_DRAWBRIDGE_GET_REJECT_TBL,&req,sizeof(rejaccTableReq)) < 0) {	
		    	dfm_perror("lookupAccRej");			 
			return ERROR;
		}
		/* Receive stats from drawbridge: */
		if(NgRecvMsg(cskt,m,sizeof(buffer),path) < 0) {		
			dfm_perror("lookupAccRej");
			return ERROR;
		}
		else
  			bcopy((rejaccTableReq *)&m->data,&req,sizeof(rejaccTableReq));		
		fprintf(theOutput, "   Packets from outside host %s are "
			"%s\n", inet_ntoa(hostaddr),
			(req.flag && req.bytes) ? "discarded" : "processed");
	} else {
		/* Send a message to drawbridge: to get stats */
		if (NgSendMsg(cskt,"drawbridge:",NGM_DRAWBRIDGE_COOKIE,
		    NGM_DRAWBRIDGE_GET_ACCEPT_TBL,&req,sizeof(rejaccTableReq)) < 0) {	
			dfm_perror("lookupAccRej");	  
			return ERROR;
		}
		/* Receive stats from drawbridge: */
		if(NgRecvMsg(cskt,m,sizeof(buffer),path) < 0) {		
			dfm_perror("lookupAccRej");
			return ERROR;
		}
		else
  			bcopy((rejaccTableReq *)&m->data,&req,sizeof(rejaccTableReq));		
		fprintf(theOutput, "   Packets from inside host %s are "
			"%s\n", inet_ntoa(hostaddr),
			req.flag ? "processed" : "discarded");
	}
	return NOERROR;
}

/*------------------------------------------------------------*/

static int
queryAcceptTable()
{
	u_int i, num;
	rejaccTableReq req;
	RejAccTableEntry *acceptTable;
	u_char buffer[2*sizeof(struct ng_mesg) + sizeof(rejaccTableReq)];
	struct ng_mesg *const m = (struct ng_mesg *) buffer;
	char path[NG_PATHLEN+1];

	req.ptr = NULL;
	req.bytes = 0;
	/* Send a message to drawbridge: to get stats */
	if (NgSendMsg(cskt,"drawbridge:",NGM_DRAWBRIDGE_COOKIE,
	    NGM_DRAWBRIDGE_GET_ACCEPT_TBL,&req,sizeof(rejaccTableReq)) < 0) {	
		dfm_perror("queryAcceptTable");
		return ERROR;
	}
	/* Receive stats from drawbridge: */
	if(NgRecvMsg(cskt,m,sizeof(buffer),path) < 0) {		
		dfm_perror("queryAcceptTable");
		return ERROR;
	}
	else
  		bcopy((rejaccTableReq *)&m->data,&req,sizeof(rejaccTableReq));	

	if (req.bytes == 0) {
		fprintf(theOutput, "   Accept address table is empty\n");
		return NOERROR;
	}

	if ((acceptTable = (RejAccTableEntry *)malloc(req.bytes)) == NULL) {
		fprintf(theErrors, "   Unable to allocate memory "
                                        "for accept table\n");
		return ERROR;
	}

	req.ptr = acceptTable;
	/* Send a message to drawbridge: to get stats */
	if (NgSendMsg(cskt,"drawbridge:",NGM_DRAWBRIDGE_COOKIE,
	    NGM_DRAWBRIDGE_GET_ACCEPT_TBL,&req,sizeof(rejaccTableReq)) < 0) {	
		dfm_perror("queryAcceptTable");	
		free(acceptTable);  
		return ERROR;
	}
	/* Receive stats from drawbridge: */
	if(NgRecvMsg(cskt,m,sizeof(buffer),path) < 0) {		
		dfm_perror("queryAcceptTable");
		free(acceptTable);
		return ERROR;
	}
	else
  		bcopy((rejaccTableReq *)&m->data,&req,sizeof(rejaccTableReq));	

	num = req.bytes / sizeof(RejAccTableEntry);
	for (i=0; i < num; i++) {
		struct in_addr a;

		if (i == 0)
			fprintf(theOutput, "   Packets from:\n");

		a.s_addr = htonl(acceptTable[i].network.s_addr);
		fprintf(theOutput, "      inside address %16s ", inet_ntoa(a));
		fprintf(theOutput, "/ %-2d  %s\n", acceptTable[i].bits,
			acceptTable[i].flag ? "process" : "discard");
	}

	fprintf(theOutput, "   All other IP addresses from the " \
		"inside will be discarded.\n");

	free(acceptTable);
	return NOERROR;
}

/*------------------------------------------------------------*/

static int
queryRejectTable()
{
	u_int i, num;
	rejaccTableReq req;
	RejAccTableEntry *rejectTable;
	u_char buffer[2*sizeof(struct ng_mesg) + sizeof(rejaccTableReq)];
	struct ng_mesg *const m = (struct ng_mesg *) buffer;
	char path[NG_PATHLEN+1];

	req.ptr = NULL;
	req.bytes = 0;
	/* Send a message to drawbridge: to get stats */
	if (NgSendMsg(cskt,"drawbridge:",NGM_DRAWBRIDGE_COOKIE,
	    NGM_DRAWBRIDGE_GET_REJECT_TBL,&req,sizeof(rejaccTableReq)) < 0) {	
		dfm_perror("queryRejectTable");	
		return ERROR;
	}
	/* Receive stats from drawbridge: */
	if(NgRecvMsg(cskt,m,sizeof(buffer),path) < 0) {		
		dfm_perror("queryRejectTable");
		return ERROR;
	}
	else
  		bcopy((rejaccTableReq *)&m->data,&req,sizeof(rejaccTableReq));	

	if (req.bytes == 0) {
		fprintf(theOutput, "   Reject address table is empty\n");
		return NOERROR;
	}

	if ((rejectTable = (RejAccTableEntry *)malloc(req.bytes)) == NULL) {
		fprintf(theErrors, "   Unable to allocate memory "
                                        "for reject table\n");
		return ERROR;
	}

	req.ptr = rejectTable;
	/* Send a message to drawbridge: to get stats */
	if (NgSendMsg(cskt,"drawbridge:",NGM_DRAWBRIDGE_COOKIE,
	    NGM_DRAWBRIDGE_GET_REJECT_TBL,&req,sizeof(rejaccTableReq)) < 0) {	
		dfm_perror("queryRejectTable");;	
		free(rejectTable);  
		return ERROR;
	}
	/* Receive stats from drawbridge: */
	if(NgRecvMsg(cskt,m,sizeof(buffer),path) < 0) {		
		dfm_perror("queryRejectTable");
		free(rejectTable);
		return ERROR;
	}
	else
  		bcopy((rejaccTableReq *)&m->data,&req,sizeof(rejaccTableReq));	


	num = req.bytes / sizeof(RejAccTableEntry);
	for (i=0; i < num; i++) {
		struct in_addr a;

		if (i == 0)
			fprintf(theOutput, "   Packets from:\n");

		a.s_addr = htonl(rejectTable[i].network.s_addr);
		fprintf(theOutput, "      outside address %16s ",inet_ntoa(a));
		fprintf(theOutput, "/ %-2d  %s\n", rejectTable[i].bits,
			rejectTable[i].flag ? "discard" : "process");
	}

	fprintf(theOutput, "   All other IP addresses from the " \
		"outside will be processed.\n");

	free(rejectTable);
	return NOERROR;
}

/*------------------------------------------------------------*/

static int
lookupOverride(char *hoststr)
{
	struct in_addr hostaddr;
	overrideTableReq req;
	u_int32_t host;
	u_char buffer[2*sizeof(struct ng_mesg) + sizeof(overrideTableReq)];
	struct ng_mesg *const m = (struct ng_mesg *) buffer;
	char path[NG_PATHLEN+1];

	if (isalpha(hoststr[0])) {
		struct hostent *h;

		if ((h = gethostbyname(hoststr))) {
			host = ntohl(*((long *) h->h_addr_list[0]));
		} else {
			fprintf(theErrors, "   Unknown host '%s'\n", hoststr);
			return ERROR;
		}
	} else
		host = inet_network(hoststr);

	req.ptr = NULL;
	req.addr = host;
	hostaddr.s_addr = htonl(host); 
	/* Send a message to drawbridge: to get stats */
	if (NgSendMsg(cskt,"drawbridge:",NGM_DRAWBRIDGE_COOKIE,
	    NGM_DRAWBRIDGE_GET_OVERRIDE_TBL,&req,sizeof(overrideTableReq)) < 0) {	
		dfm_perror("lookupOverride");
		return ERROR;
	}
	/* Receive stats from drawbridge: */
	if(NgRecvMsg(cskt,m,sizeof(buffer),path) < 0) {		
		dfm_perror("lookupOverride");
		return ERROR;
	}
	else
  		bcopy((overrideTableReq *)&m->data,&req,sizeof(overrideTableReq));	
	if (req.index != 0) {
		fprintf(theOutput, "   Host %s is using override table index "
				   "%d.\n", inet_ntoa(hostaddr), req.index);
	} else {
		fprintf(theOutput, "   Host %s is not in the override "
				   "table.\n", inet_ntoa(hostaddr));
	}

	return NOERROR;
}

/*------------------------------------------------------------*/

int
queryOverrideTable()
{
	u_int i, num;
	overrideTableReq req;
	OverrideTableEntry *table;
	u_char buffer[2*sizeof(struct ng_mesg) + sizeof(overrideTableReq)];
	struct ng_mesg *const m = (struct ng_mesg *) buffer;
	char path[NG_PATHLEN+1];

	req.ptr = NULL;
	req.bytes = 0;
	/* Send a message to drawbridge: to get stats */
	if (NgSendMsg(cskt,"drawbridge:",NGM_DRAWBRIDGE_COOKIE,
	    NGM_DRAWBRIDGE_GET_OVERRIDE_TBL,&req,sizeof(overrideTableReq)) < 0) {	
		dfm_perror("queryOverrrideTable");
		return ERROR;
	}
	/* Receive stats from drawbridge: */
	if(NgRecvMsg(cskt,m,sizeof(buffer),path) < 0) {		
		dfm_perror("queryOverrideTable");
		return ERROR;
	}
	else
  		bcopy((overrideTableReq *)&m->data,&req,sizeof(overrideTableReq));	

	if (req.bytes == 0) {
		fprintf(theOutput, "   Override address table is empty\n");
		return NOERROR;
	}

	if ((table = (OverrideTableEntry *)malloc(req.bytes)) == NULL) {
		fprintf(theErrors, "   Unable to allocate memory "
                                        "for override table\n");
		return ERROR;
	}

	req.ptr = table;
	/* Send a message to drawbridge: to get stats */
	if (NgSendMsg(cskt,"drawbridge:",NGM_DRAWBRIDGE_COOKIE,
	    NGM_DRAWBRIDGE_GET_OVERRIDE_TBL,&req,sizeof(overrideTableReq)) < 0) {	
		dfm_perror("queryOverrideTable");	
		free(table);
		return ERROR;
	}
	/* Receive stats from drawbridge: */
	if(NgRecvMsg(cskt,m,sizeof(buffer),path) < 0) {		
		dfm_perror("queryOverrideTable");
		free(table);
		return ERROR;
	}
	else
  		bcopy((overrideTableReq *)&m->data,&req,sizeof(overrideTableReq));	

	num = req.bytes / sizeof(OverrideTableEntry);
	for (i=0; i < num; i++) {
		register u_int j;
		struct in_addr a;

		if (i == 0)
			fprintf(theOutput, "   Allow all outgoing " \
						"packets to:\n");

		a.s_addr = htonl(table[i].network.s_addr);
		fprintf(theOutput, "      address %16s ", inet_ntoa(a));
		fprintf(theOutput, "/ %-2d  ports ", table[i].bits);
		for (j=1; j < LIST_SIZE; j++) {
			if (j > table[i].port[0].begin)
				break;
			if (j>1) fprintf(theOutput, "%43s", " ");
			fprintf(theOutput, "%5d-%-5d\n",
				table[i].port[j].begin,
				table[i].port[j].end);
		}
	}

	fprintf(theOutput, "   All other IP addresses from the " \
		"outside will be processed.\n");

	free(table);
	return NOERROR;
}

/*------------------------------------------------------------*/

static char
print_line(PortListEntry *table, u_short i)
{
	char flag = 0;

	if (i <= table[0].begin) {
		fprintf(theOutput, "   %5d-%-5d",
			table[i].begin, table[i].end);
		flag = 1;
	} else
		fprintf(theOutput, "%14s"," ");

	return flag;
}

/*------------------------------------------------------------*/

static int
queryTCPresetTable()
{
	u_short i;
	char flag;
	PortListEntry t_rst_in[LIST_SIZE];
	PortListEntry t_rst_out[LIST_SIZE];
	u_int bytes;

	bytes = LIST_TABLE_SIZE;

	if (portListRequest(READ, T_RST_IN, 0, t_rst_in, bytes))
		return ERROR;
	if (portListRequest(READ, T_RST_OUT, 0, t_rst_out, bytes))
		return ERROR;

	fprintf(theOutput, "     TCP dst       TCP dst\n");
	fprintf(theOutput, "     port in       port out\n"); 
	fprintf(theOutput, "   -----------   -----------\n");

	for (i=1,flag=1; i<LIST_SIZE && flag; i++) {
		flag = 0;
		if (i <= t_rst_in[0].begin) {
			fprintf(theOutput, "   %5d-%-5d",
			  t_rst_in[i].begin, t_rst_in[i].end);
			flag = 1;
		} else
			fprintf(theOutput, "%14s"," ");

		if (i <= t_rst_out[0].begin) {
			fprintf(theOutput, "   %5d-%-5d\n",
			  t_rst_out[i].begin, t_rst_out[i].end);
			flag = 1;
		} else
			fprintf(theOutput, "\n");
	}

	return NOERROR;
}

/*------------------------------------------------------------*/

static int
queryClassTable(u_short class, char *tok)
{
	u_short i;
	char flag;
	PortListEntry t_dst_in[LIST_SIZE];
	PortListEntry t_dst_out[LIST_SIZE];
	PortListEntry t_src_in[LIST_SIZE];
	PortListEntry t_src_out[LIST_SIZE];
	PortListEntry u_dst_in[LIST_SIZE];
	PortListEntry u_dst_out[LIST_SIZE];
	PortListEntry u_src_in[LIST_SIZE];
	PortListEntry u_src_out[LIST_SIZE];
	PortListEntry i_typ_in[LIST_SIZE];
	PortListEntry i_typ_out[LIST_SIZE];
	u_int bytes;

	bytes = LIST_TABLE_SIZE;

	if (portListRequest(READ, T_DST_IN, class, t_dst_in, bytes))
		return ERROR;
	if (portListRequest(READ, T_DST_OUT, class, t_dst_out, bytes))
		return ERROR;
	if (portListRequest(READ, T_SRC_IN, class, t_src_in, bytes))
		return ERROR;
	if (portListRequest(READ, T_SRC_OUT, class, t_src_out, bytes))
		return ERROR;
	if (portListRequest(READ, U_DST_IN, class, u_dst_in, bytes))
		return ERROR;
	if (portListRequest(READ, U_DST_OUT, class, u_dst_out, bytes))
		return ERROR;
	if (portListRequest(READ, U_SRC_IN, class, u_src_in, bytes))
		return ERROR;
	if (portListRequest(READ, U_SRC_OUT, class, u_src_out, bytes))
		return ERROR;
	if (portListRequest(READ, I_TYP_IN, class, i_typ_in, bytes))
		return ERROR;
	if (portListRequest(READ, I_TYP_OUT, class, i_typ_out, bytes))
		return ERROR;

	if (tok && tolower(*tok) == 'w') {
		fprintf(theOutput,
		"     TCP dst       TCP src       UDP dst       "
		"UDP src       ICMP dst      TCP dst       TCP src       "
		"UDP dst       UDP src       ICMP dst\n");
		fprintf(theOutput,
		"     port in       port in       port in       "
		"port in       type in       port out      port out      "
		"port out      port out      type out\n");
		fprintf(theOutput,
		"   -----------   -----------   -----------   "
		"-----------   -----------   -----------   -----------   "
		"-----------   -----------   -----------\n");
		for (i=1,flag=1; i<LIST_SIZE && flag; i++) {
			flag=0;
			if (print_line(t_dst_in, i)) flag = 1;
			if (print_line(t_src_in, i)) flag = 1;
			if (print_line(u_dst_in, i)) flag = 1;
			if (print_line(u_src_in, i)) flag = 1;
			if (print_line(i_typ_in, i)) flag = 1;
			if (print_line(t_dst_out, i)) flag = 1;
			if (print_line(t_src_out, i)) flag = 1;
			if (print_line(u_dst_out, i)) flag = 1;
			if (print_line(u_src_out, i)) flag = 1;
			if (print_line(i_typ_out, i)) flag = 1;
			fprintf(theOutput, "\n");
		}
	} else {
		fprintf(theOutput,
		"     TCP dst       TCP src       UDP dst" \
		"       UDP src       ICMP dst\n");
		fprintf(theOutput,
		"     port in       port in       port in" \
		"       port in       type in\n");
		fprintf(theOutput,
		"   -----------   -----------   -----------" \
		"   -----------   -----------\n");
		for (i=1,flag=1; i<LIST_SIZE && flag; i++) {
			flag=0;
			if (print_line(t_dst_in, i)) flag = 1;
			if (print_line(t_src_in, i)) flag = 1;
			if (print_line(u_dst_in, i)) flag = 1;
			if (print_line(u_src_in, i)) flag = 1;
			if (print_line(i_typ_in, i)) flag = 1;
			fprintf(theOutput, "\n");
		}
		fprintf(theOutput,
		"     TCP dst       TCP src       UDP dst" \
		"       UDP src       ICMP dst\n");
		fprintf(theOutput,
		"     port out      port out      port out" \
		"      port out      type out\n");
		fprintf(theOutput,
		"   -----------   -----------   -----------" \
		"   -----------   -----------\n");
		for (i=1,flag=1; i<LIST_SIZE && flag; i++) {
			flag=0;
			if (print_line(t_dst_out, i)) flag = 1;
			if (print_line(t_src_out, i)) flag = 1;
			if (print_line(u_dst_out, i)) flag = 1;
			if (print_line(u_src_out, i)) flag = 1;
			if (print_line(i_typ_out, i)) flag = 1;
			fprintf(theOutput, "\n");
		}
	}

	return NOERROR;
}

/*------------------------------------------------------------*/

static int
queryNetworkTable(char *hoststr, char cflag, char *tok)
{
	struct in_addr hostaddr;
	addrTreeReq u;
	u_long host;
	u_char buffer[2*sizeof(struct ng_mesg) + sizeof(addrTreeReq)];
	struct ng_mesg *const m = (struct ng_mesg *) buffer;
	char path[NG_PATHLEN+1];

	if (isalpha(hoststr[0])) {
		struct hostent *h;

		if ((h = gethostbyname(hoststr))) {
			host = ntohl(*((long *) h->h_addr_list[0]));
		} else {
			fprintf(theErrors, "   Unknown host '%s'\n", hoststr);
			return ERROR;
		}
	} else
		host = inet_network(hoststr);

	u.num = 1;
	u.newtable = 0;
	u.addr[0].s_addr = host;
	u.bits[0] = 32;
	/* Send a message to drawbridge: to get stats */
	if (NgSendMsg(cskt,"drawbridge:",NGM_DRAWBRIDGE_COOKIE,
	    NGM_DRAWBRIDGE_GET_ADDRESS_TBL,&u,sizeof(addrTreeReq)) < 0) {	
		dfm_perror("queryNetworkTable");	  
		return ERROR;
	}
	/* Receive stats from drawbridge: */
	if(NgRecvMsg(cskt,m,sizeof(buffer),path) < 0) {		
		dfm_perror("queryNetworkTable");
		return ERROR;
	}
	else
  		bcopy((addrTreeReq *)&m->data,&u,sizeof(addrTreeReq));
	hostaddr.s_addr = htonl(host);
	fprintf(theOutput, "   Host %s is using class %d\n",
		inet_ntoa(hostaddr), u.class[0]);

	if (cflag) {
		fprintf(theOutput, "\n");
		queryClassTable(u.class[0], tok);
	}

	return NOERROR;
}

/*------------------------------------------------------------*/

static void
showStats()
{
	struct ng_drawbridge_stats theStats;
	Status theStatus;
	u_long upTime, upDays, upHours, upMinutes, upSeconds;
	u_char buffer[2*sizeof(struct ng_mesg) + sizeof(struct ng_drawbridge_stats)];
	struct ng_mesg *const m = (struct ng_mesg *) buffer;
	u_char buffer2[2*sizeof(struct ng_mesg) + sizeof(Status)];
	struct ng_mesg *const m2 = (struct ng_mesg *) buffer2;
	char path[NG_PATHLEN+1];
	
	/* Send a message to drawbridge: to get stats */
	if (NgSendMsg(cskt,"drawbridge:",NGM_DRAWBRIDGE_COOKIE,
	    NGM_DRAWBRIDGE_GET_STATS,NULL,0) < 0) {	
		dfm_perror("showStats");  
		return;
	}
	/* Receive stats from drawbridge: */
	if(NgRecvMsg(cskt,m,sizeof(buffer),path) < 0) {		
		dfm_perror("showStats");
		return;
	}
	else
  		bcopy((struct ng_drawbridge_stats *)&m->data,&theStats,
		       sizeof(struct ng_drawbridge_stats));

	fprintf(theOutput, "\nPackets%33s%34s\n", "Inside", "Outside");
	fprintf(theOutput, "-------%33s%34s\n", "------", "-------");
	fprintf(theOutput, "Filtered %31s", fmt_big(theStats.p_insideFiltered));
	fprintf(theOutput, "%34s\n", fmt_big(theStats.p_outsideFiltered));
	fprintf(theOutput, "Received    %28s", fmt_big(theStats.p_insideRx));
	fprintf(theOutput, "%34s\n", fmt_big(theStats.p_outsideRx));
	fprintf(theOutput, "Transmitted %28s", fmt_big(theStats.p_insideTx));
	fprintf(theOutput, "%34s\n\n", fmt_big(theStats.p_outsideTx));

	fprintf(theOutput, "Bytes%35s%34s\n", "Inside", "Outside");
	fprintf(theOutput, "-----%35s%34s\n", "------", "-------");
	fprintf(theOutput, "Filtered %31s", fmt_big(theStats.b_insideFiltered));
	fprintf(theOutput, "%34s\n", fmt_big(theStats.b_outsideFiltered));
	fprintf(theOutput, "Received    %28s", fmt_big(theStats.b_insideRx));
	fprintf(theOutput, "%34s\n", fmt_big(theStats.b_outsideRx));
	fprintf(theOutput, "Transmitted %28s", fmt_big(theStats.b_insideTx));
	fprintf(theOutput, "%34s\n\n", fmt_big(theStats.b_outsideTx));

	fprintf(theOutput, "Bridge table entries:%11s\n\n",
                fmt_small(theStats.bridgeEntries, NO));


			
	
	/* Send a message to drawbridge: to get status */
	if (NgSendMsg(cskt,"drawbridge:",NGM_DRAWBRIDGE_COOKIE,
	    NGM_DRAWBRIDGE_GET_STATUS,NULL,0) < 0) {	
	    	dfm_perror("showStats");	  
		return;
	}
	/* Receive stats from drawbridge: */
	if(NgRecvMsg(cskt,m2,sizeof(buffer),path) < 0) {
		dfm_perror("showStats");		
		return;
	}
	else
  		bcopy((Status *)&m2->data,&theStatus,sizeof(Status));

	if (theStatus.running) {
		upDays = upHours = upMinutes = upSeconds = 0;
		upTime = time(NULL) - theStatus.startTime.tv_sec;
		if (upTime) {
			upDays = upTime / (24UL * 60UL * 60UL);
			upTime -= upDays * 24UL * 60UL * 60UL;
			upHours = upTime / (60 * 60);
			upTime -= upHours * 60 * 60;
			upMinutes = upTime / 60;
			upSeconds = upTime - upMinutes * 60;
		}
		fprintf(theOutput, "Uptime: %lu days %lu hours %lu minutes %lu "
			"seconds\n\n", upDays, upHours, upMinutes, upSeconds);
	} else
		fprintf(theOutput, "Drawbridge is not running.\n\n");
}

/*------------------------------------------------------------*/

void
handleShow()
{
	int result;
	char *currTok;
	u_char buffer[2*sizeof(struct ng_mesg) + sizeof(struct ng_drawbridge_flags)];
	u_char buffer2[2*sizeof(struct ng_mesg) + sizeof(bridgeReq)];
	struct ng_mesg *const m = (struct ng_mesg *) buffer;
	struct ng_mesg *const m2 = (struct ng_mesg *) buffer2;
	char path[NG_PATHLEN+1];


	currTok = strtok(NULL, DELIMITERS);
	result = findtoken(currTok);

	switch (result) {

	case T_HOST:
		currTok = strtok(NULL, DELIMITERS);
		if (!currTok) {
			fprintf(theErrors, "   Missing host address/name\n");
			break;
		}
		if (*currTok == '?')
			usageShow(VERBOSE);
		else
			queryNetworkTable(currTok, NO, NULL);
		break;

	case T_CLASS:
		currTok = strtok(NULL, DELIMITERS);
		if (!currTok) {
			fprintf(theErrors, "   Missing class number\n");
			break;
		}
		if (strIsNumeric(currTok))
			queryClassTable(atoi(currTok),
				strtok(NULL, DELIMITERS));
		else
			if (*currTok == '?')
				usageShow(VERBOSE);
			else
				queryNetworkTable(currTok, YES,
					strtok(NULL, DELIMITERS));
		break;

	case T_TCPRESET:
		queryTCPresetTable();
		break;

	case T_ACCEPT:
		currTok = strtok(NULL, DELIMITERS);
		if (!currTok)
			queryAcceptTable();
		else
			if (*currTok == '?')
				usageShow(VERBOSE);
			else
				lookupAccRej(T_ACCTBL, currTok);
		break;

	case T_REJECT:
		currTok = strtok(NULL, DELIMITERS);
		if (!currTok)
			queryRejectTable();
		else
			if (*currTok == '?')
				usageShow(VERBOSE);
			else
				lookupAccRej(T_REJTBL, currTok);
		break;

	case T_OVERRIDE:
		currTok = strtok(NULL, DELIMITERS);
		if (!currTok)
			queryOverrideTable();
		else
			if (*currTok == '?')
				usageShow(VERBOSE);
			else
				lookupOverride(currTok);
		break;

	case T_FLAGS: {
		struct ng_drawbridge_flags flags;
		
		/* Send a message to drawbridge: to get stats */
		if (NgSendMsg(cskt,"drawbridge:",NGM_DRAWBRIDGE_COOKIE,
	   	    NGM_DRAWBRIDGE_GET_FLAGS,NULL,0) < 0) {	
			dfm_perror("handleShow");	  
			break;
		}
		/* Receive stats from drawbridge: */
		if(NgRecvMsg(cskt,m,sizeof(buffer),path) < 0) {		
			dfm_perror("handleShow");
			break;
		}
		else
  			bcopy((struct ng_drawbridge_flags *)&m->data,&flags,
		              sizeof(struct ng_drawbridge_flags));

		fprintf(theOutput, "   Multicast      = %s\n",
			flags.discardMulticast ? "discard" : "forward");
		fprintf(theOutput, "   NonIp          = %s\n",
			flags.discardNonIp ? "discard" : "forward");
		fprintf(theOutput, "   OtherIp        = %s\n",
			flags.discardOtherIp ? "discard" : "forward");
		fprintf(theOutput, "   SuspectOffset  = %s\n",
			flags.discardSuspectOffset ? "discard" : "forward");
		fprintf(theOutput, "   FragmentedICMP = %s\n",
			flags.discardFragmentedICMP ? "discard" : "forward");
		fprintf(theOutput, "   AttackICMP     = %s\n",
			flags.discardAttackICMP ? "discard" : "forward");
		break;
	}

	case T_STATS:
		showStats();
		break;
		
	case T_BRIDGE: {
		int i;
		bridgeReq breq;
		for (i=0; i<MAX_BRIDGE_ADDRESSES; i++) {
			breq.t.valid = NO;
			breq.index = i;
			
			/* Send a message to drawbridge: to get bridge entries */
			if (NgSendMsg(cskt,"drawbridge:",NGM_DRAWBRIDGE_COOKIE,
		   	    NGM_DRAWBRIDGE_GET_BRIDGE_ENTRY,&breq,sizeof(bridgeReq)) < 0) {	
				dfm_perror("handleShow");	  
				break;
			}
			/* Receive stats from drawbridge: */
			if(NgRecvMsg(cskt,m2,sizeof(buffer2),path) < 0) {		
				dfm_perror("handleShow");
				break;
			}
			else 
  				bcopy((struct bridgeReq *)&m2->data,&breq,
		        	      sizeof(bridgeReq));
				      
			if (breq.t.valid == YES) {
				fprintf(theOutput, "   " \
					"%02x:%02x:%02x:%02x:%02x:%02x   %s\n",
					breq.t.address.bytes[0],
					breq.t.address.bytes[1],
					breq.t.address.bytes[2],
					breq.t.address.bytes[3],
					breq.t.address.bytes[4],
					breq.t.address.bytes[5],
					(breq.t.interface == DB_INSIDE_IF) ?
						"inside" : "outside");
			}
		}
		break;
	}

	default:
		usageShow(TERSE);
	}
}

/*------------------------------------------------------------*/
