#!/bin/sh

#
# Drawbridge Startup Script 
# $Id: drawbridge.sh,v 1.1.1.1 2004/01/06 21:09:54 daryl Exp $
# 

#
# Define interface IDs
#

# If you are not using a mirror interface, set mirror = "" below.
outside="fxp1"
inside="fxp2"
mirror="fxp3"

# Shouldn't need to change anything below this line. 

#
# Define paths
#
drawbridge="/usr/local/drawbridge"
ruleset="/usr/local/drawbridge/rules/db_filters"

case "$1" in
start)

	#
	# Check to see if Drawbridge is already loaded 
	# or is present in the kernel
	#
	kldstat -v | grep ng_drawbridge > /dev/null
	if [ $? -eq 0 ]
	then
		echo "Drawbridge module already loaded."
		exit $?
	fi	

	echo ""
 	echo "Initializing Drawbridge..."	

	#
	# Enable device polling
	#
	sysctl -N kern | grep polling.enable > /dev/null
	if [ $? -eq 0 ]
	then
		echo "Enabling device polling."
		sysctl kern.polling.enable=1 > /dev/null	
	fi

	#
	# Check to see if netgraph and ng_ether are loaded
	# or are present in the kernel
	#
	kldstat -v | grep netgraph > /dev/null 
	if [ $? -ne 0 ] 
	then
  		echo "Drawbridge: Loading netgraph kernel module."  
  		kldload netgraph > /dev/null
		if [ $? -ne 0 ]
		then
			echo "An error occurred loading the netgraph kernel module."
			exit $?
		fi
	fi

	kldstat -v | grep ng_ether > /dev/null
	if [ $? -ne 0 ] 
	then
  		echo "Drawbridge: Loading ng_ether kernel module."
  		kldload ng_ether > /dev/null
		if [ $? -ne 0 ]
		then
			echo "An error occurred loading the ng_ether kernel module."
			exit $?
		fi
	fi

	#
	# Bring up network interfaces
	#
	ifconfig $outside up &2> /dev/null
	if [ $? -ne 0 ] 
	then
	 	echo "An error occurred bringing up outside interface $outside."
		exit $?
	fi
	ifconfig $inside up &2> /dev/null
	if [ $? -ne 0 ] 
	then
	 	echo "An error occurred bringing up inside interface $inside."
		exit $?
	fi
	if [ $mirror ]
	then
		ifconfig $mirror up &2> /dev/null
		if [ $? -ne 0 ] 
		then
		 	echo "An error occurred bringing up mirror interface $mirror."
			exit $?
		fi
	fi

	#
	# Load Drawbridge node and connect to outside interface 
	#
	ngctl mkpeer $outside: drawbridge lower outside
	if [ $? -ne 0 ] 
	then
  		echo "Drawbridge: Error loading and connecting Drawbridge node."
  		exit $?
	fi

	#
	# Name Drawbridge node
	#
	ngctl name $outside:lower drawbridge
	if [ $? -ne 0 ] 
	then
  		echo "Drawbridge: Error naming Drawbrige node."
  		exit $?
	fi

	#
	# Connect Drawbridge to inside interface
	#
	ngctl connect drawbridge: $inside: inside lower
	if [ $? -ne 0 ] 
	then 
  		echo "Drawbridge: Error connecting Drawbridge node to inside interface."
  		exit $?
	fi

	#
	# Connect Drawbridge to mirror interface
	#
	if [ $mirror ]
	then
		ngctl connect drawbridge: $mirror: mirror lower
		if [ $? -ne 0 ] 
		then
			echo "Drawbridge: Error connecting Drawbridge node to mirror interface."
			exit $?
		fi
	fi

	#
	# Complete Drawbridge initialization
	#
	if [ -f $ruleset ]
	then
		echo "Drawbridge: Loading Drawbridge ruleset..."
		$drawbridge/sbin/dbmgr load fil $drawbridge/rules/db_filters
 	else
		echo "Drawbridge: Ruleset file not found.  Using default ruleset."
	fi 
	
	#
	# Place outside and inside interfaces in promiscuous mode and squash autosrc 
	#
	ngctl msg $outside: setautosrc 0 
	if [ $? -ne 0 ] 
	then
  		echo "Drawbridge: Error squashing autosrc on outside interface."
  		exit $?
	fi
	
	ngctl msg $inside: setautosrc 0 
	if [ $? -ne 0 ] 
	then
  		echo "Drawbridge: Error squashing autosrc on inside interface." 
  		exit $?
	fi
	
	ngctl msg $outside: setpromisc 1
	if [ $? -ne 0 ] 
	then
  		echo "Drawbridge: Error placing outside interface in promiscuous mode."
  		exit $?
	fi
	
	ngctl msg $inside: setpromisc 1
	if [ $? -ne 0 ] 
	then
  		echo "Drawbridge: Error placing inside interface in promiscuous mode." 
  		exit $?
	fi
	
	echo "Drawbridge initialization complete."
	;;
stop)
	#
	# Check to see if Drawbridge is already running
	#
	kldstat | grep ng_drawbridge > /dev/null
	if [ $? -ne 0 ]
	then
		echo "Drawbridge module not loaded."
		exit $?
	fi	

	#
	# Shutdown Drawbridge
	#
	ngctl shutdown drawbridge: 
 	if [ $? -ne 0 ] 
	then
		echo "Drawbridge: Error shutting down Drawbridge."
		exit $?
	fi

	#	
	# Shutdown network interfaces
	#
	ngctl shutdown $outside:
	if [ $? -ne 0 ] 
	then
		echo "Drawbridge: Error shutting down outside interface."
		exit $?
	fi

 	ngctl shutdown $inside:	
	if [ $? -ne 0 ] 	
	then
		echo "Drawbridge: Error shutting down inside interface."
		exit $?
	fi

	if [ $mirror ]
	then
		ngctl shutdown $mirror:
		if [ $? -ne 0 ] 
		then
			echo "Drawbridge: Error shutting down mirror interface."
			exit $?
		fi
	fi

	#
	# Unload Drawbridge module
	#
	kldunload ng_drawbridge
	if [ $? -ne 0 ]
	then
		echo "Drawbridge: Error unloading Drawbridge module."
		exit $?
	fi

	#
	# Shutdown complete
	#
	echo "Drawbridge shutdown complete."
	;;	
*)
	echo "Usage: `basename $0` {start|stop}" >&2
	;;
esac

