/*
 * Copyright (c) 1993-2003
 *      Texas A&M University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Texas A&M University
 *      and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Developers:
 *     Russell Neeper, David K. Hess, Douglas Lee Schales, David R. Safford,
 *     Wm. Daryl Hawkins
 *
 * $Id: ng_drawbridge.c,v 1.1.1.1 2004/01/06 21:09:54 daryl Exp $
 *
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/errno.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/socket.h>
#include <sys/mbuf.h>
#include <sys/syslog.h>

#include <net/if.h>
#include <net/if_types.h>
#include <net/if_llc.h>

#include <netinet/in.h>
#include <netinet/in_var.h>
#include <netinet/if_ether.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>

#include <machine/in_cksum.h>
#include <machine/stdarg.h>

#include <netgraph/ng_message.h>
#include <netgraph/netgraph.h>
#include <netgraph/ng_parse.h>
#include <netgraph/ng_drawbridge.h>

/* Drawbridge variables */
pointerStruct active, new;
int logFacility = DEFAULT_SYSL_FACILITY;
u_int32_t logMask = DEFAULT_SYSL_MASK;

/* MAC address hash table for bridge code */
/* Addresses are stored in network byte order */
HashEntry hashTable [MAX_BRIDGE_ADDRESSES];

/* Per hook info */
struct hookinfo {
	hook_p			hook;	
};

/* Per node info */
struct privdata {
	node_p				node;
	struct hookinfo	 		outside;	
	struct hookinfo			inside;
	struct hookinfo			mirror;
	struct ng_drawbridge_stats 	stats;
	struct ng_drawbridge_flags	flags;
	Status				DBstatus;
};
typedef struct privdata *sc_p;

/* Netgraph methods */
static ng_constructor_t	ng_drawbridge_constructor;
static ng_rcvmsg_t	ng_drawbridge_rcvmsg;
static ng_shutdown_t	ng_drawbridge_rmnode;
static ng_newhook_t	ng_drawbridge_newhook;
static ng_rcvdata_t	ng_drawbridge_rcvdata;
static ng_disconnect_t	ng_drawbridge_disconnect;

/* Parse type for struct ng_drawbridge_stats */
static const struct ng_parse_struct_field ng_drawbridge_stats_type_fields[]
	= NG_DRAWBRIDGE_STATS_TYPE_INFO;
static const struct ng_parse_type ng_drawbridge_stats_type = {
	&ng_parse_struct_type,
	&ng_drawbridge_stats_type_fields
};

/* Parse type for struct ng_drawbridge_flags */
static const struct ng_parse_struct_field ng_drawbridge_flags_type_fields[]
	= NG_DRAWBRIDGE_FLAGS_TYPE_INFO;
static const struct ng_parse_type ng_drawbridge_flags_type = {
	&ng_parse_struct_type,
	&ng_drawbridge_flags_type_fields 
};

/* List of commands and how to convert arguments to/from ASCII */
static const struct ng_cmdlist ng_drawbridge_cmds[] = {
	{
	  NGM_DRAWBRIDGE_COOKIE,
	  NGM_DRAWBRIDGE_GET_STATS,
	  "get_stats", 
	  NULL,
	  &ng_drawbridge_stats_type
	},
	{
	  NGM_DRAWBRIDGE_COOKIE,
	  NGM_DRAWBRIDGE_CLR_STATS,
	  "clrstats",
	  NULL,
	  NULL
	},
	{
	  NGM_DRAWBRIDGE_COOKIE,
	  NGM_DRAWBRIDGE_GET_CLR_STATS,
	  "get_clrstats",
	  NULL,
	  &ng_drawbridge_stats_type
	},
	{
	  NGM_DRAWBRIDGE_COOKIE,
	  NGM_DRAWBRIDGE_SET_FLAGS,
	  "set_flags",
	  &ng_drawbridge_flags_type,
	  NULL	  
	},
	{
	  NGM_DRAWBRIDGE_COOKIE,
	  NGM_DRAWBRIDGE_GET_FLAGS,
	  "get_flags",
	  NULL,
	  &ng_drawbridge_flags_type
	},
	{
	  NGM_DRAWBRIDGE_COOKIE,
	  NGM_DRAWBRIDGE_GET_VERSION,
	  "get_version",
	  NULL,
	  &ng_parse_string_type
	}, 
	/* 
	 * From this point on, we don't care about conversion to ASCII 
	 * The following commands are binary only 
	 */
	{
	  NGM_DRAWBRIDGE_COOKIE,
	  NGM_DRAWBRIDGE_GET_STATUS,
	  NULL,
	  NULL,
	  NULL
	},
	{
	  NGM_DRAWBRIDGE_COOKIE,
	  NGM_DRAWBRIDGE_SET_ADDRESS_TBL,
	  NULL,
	  NULL,
	  NULL
	},
	{
	  NGM_DRAWBRIDGE_COOKIE,
	  NGM_DRAWBRIDGE_GET_ADDRESS_TBL,
	  NULL,
	  NULL,
	  NULL
	},
	{
	  NGM_DRAWBRIDGE_COOKIE,
	  NGM_DRAWBRIDGE_SET_CLASS_TBL,
	  NULL,
	  NULL,
	  NULL
	},
	{
	  NGM_DRAWBRIDGE_COOKIE,
	  NGM_DRAWBRIDGE_GET_CLASS_TBL,
	  NULL,
	  NULL,
	  NULL
	},
	{
	  NGM_DRAWBRIDGE_COOKIE,
	  NGM_DRAWBRIDGE_SET_ACCEPT_TBL,
	  NULL,
	  NULL,
	  NULL
	},
	{
	  NGM_DRAWBRIDGE_COOKIE,
	  NGM_DRAWBRIDGE_GET_ACCEPT_TBL,
	  NULL,
	  NULL,
	  NULL
	},
	{
	  NGM_DRAWBRIDGE_COOKIE,
	  NGM_DRAWBRIDGE_SET_REJECT_TBL,
	  NULL,
	  NULL,
	  NULL
	},
	{
	  NGM_DRAWBRIDGE_COOKIE,
	  NGM_DRAWBRIDGE_GET_REJECT_TBL,
	  NULL,
	  NULL,
	  NULL
	},
	{
	  NGM_DRAWBRIDGE_COOKIE,
	  NGM_DRAWBRIDGE_SET_OVERRIDE_TBL,
	  NULL,
	  NULL,
	  NULL
	},
	{
	  NGM_DRAWBRIDGE_COOKIE,
	  NGM_DRAWBRIDGE_GET_OVERRIDE_TBL,
	  NULL,
	  NULL,
	  NULL
	},	
	{
	  NGM_DRAWBRIDGE_COOKIE,
	  NGM_DRAWBRIDGE_MANAGE_TABLES,
	  NULL,
	  NULL,
	  NULL
	},
	{
	  NGM_DRAWBRIDGE_COOKIE,
	  NGM_DRAWBRIDGE_GET_BRIDGE_ENTRY,
	  NULL,
	  NULL,
	  NULL
	},
	{
	  NGM_DRAWBRIDGE_COOKIE,
	  NGM_DRAWBRIDGE_CLR_BRIDGE_TBL,
	  NULL,
	  NULL,
	  NULL
	},
	{ 0 }
};

static SyslogMessageEntry syslogMessages[] = {
	{ "unknown event",				LOG_WARNING },
	{ "cold start",					LOG_INFO    },
	{ "incoming class D:",				LOG_WARNING },
	{ "outgoing class D:",				LOG_WARNING },
	{ "incoming port:",				LOG_WARNING },
	{ "outgoing port:",				LOG_WARNING },
	{ "incoming type:",				LOG_WARNING },
	{ "outgoing type:",				LOG_WARNING },
	{ "incoming via reject table:",			LOG_WARNING },
	{ "outgoing via accept table:",			LOG_WARNING },
	{ "outgoing via override table:",		LOG_WARNING },
	{ "incoming header too short:",			LOG_WARNING },
	{ "outgoing header too short:",			LOG_WARNING },
	{ "incoming D-O-S attack:",			LOG_WARNING },
	{ "outgoing D-O-S attack:",			LOG_WARNING },
	{ "incoming IP:",				LOG_NOTICE  },
	{ "outgoing IP:",				LOG_WARNING },
	{ "incoming fragment with IP offset == 1:",	LOG_WARNING },
	{ "outgoing fragment with IP offset == 1:",	LOG_WARNING },
	{ "incoming fragment:",				LOG_WARNING },
	{ "outgoing fragment:",				LOG_WARNING },
	{ "incoming MAC layer protocol:",		LOG_INFO    },
	{ "outgoing MAC layer protocol:",		LOG_WARNING },
	{ NULL } /* Marks the end of the messages. */
};

static const char *syslogDosTypes[] = { "\"Smurf\"", "\"Pong\"" };

/* Netgraph type descriptor */
static struct ng_type ng_drawbridge_typestruct = {
	NG_VERSION,
	NG_DRAWBRIDGE_NODE_TYPE,
	NULL,
	ng_drawbridge_constructor,
	ng_drawbridge_rcvmsg,
	ng_drawbridge_rmnode,
	ng_drawbridge_newhook,
	NULL,
	NULL,
	ng_drawbridge_rcvdata,
	ng_drawbridge_rcvdata,
	ng_drawbridge_disconnect,
	ng_drawbridge_cmds
};
NETGRAPH_INIT(drawbridge, &ng_drawbridge_typestruct);

int
drawbridge_test_pkt(const sc_p sc, struct hookinfo *const hinfo,struct mbuf *m);

void
drawbridge_response_pkt(struct mbuf *r, struct mbuf *m);

static int
initTables(pointerStruct *p, u_int16_t classes);

static int
bridgeRequest(int, bridgeReq *);

static int
manageTablesRequest(manageTablesReq *u);

static int
addrTreeRequest(int rw, addrTreeReq *u);

static int
portListRequest(int rw, portListReq *u, pointerStruct *p);

static int
rejaccTableRequest(int rw, int id, rejaccTableReq *u, pointerStruct *p);

static int
overrideTableRequest(int rw, overrideTableReq *u, pointerStruct *p);

static void
syslogMessage(u_int eventNo,...);

/*
 * Node constructor
 */
static int
ng_drawbridge_constructor(node_p *nodep)
{
	sc_p privdata;
	int error = 0;
	struct timeval curTime;
			
	MALLOC(privdata, sc_p, sizeof(*privdata), M_NETGRAPH, M_NOWAIT);
	if (privdata == NULL)
		return (ENOMEM);
	bzero(privdata, sizeof(*privdata));

	if ((error = ng_make_node_common(&ng_drawbridge_typestruct, nodep))) {
		FREE(privdata, M_NETGRAPH);
		return (error);
	}
	(*nodep)->private = privdata;
	privdata->node = *nodep;
	privdata->DBstatus.running = NO;
	
	/* Initialize Drawbridge */	
	bzero(&active, sizeof(pointerStruct));
	bzero(&new, sizeof(pointerStruct));

	/* Clear the bridge table */
	bzero(hashTable, sizeof(hashTable));
	
	/* Initialize default filters */
	initTables(&active, 1);
	
	/* Apparently all went well... or at least it looks that way. */	
	getmicrotime(&curTime);
	privdata->DBstatus.startTime = curTime;
	privdata->DBstatus.running = YES;
	
	return (0);
}

/*
 * Add module hooks
 */
static int
ng_drawbridge_newhook(node_p node, hook_p hook, const char *name)
{
	const sc_p sc = node->private;

	if (strcmp(name, NG_DRAWBRIDGE_HOOK_INSIDE) == 0) {
		sc->inside.hook = hook;
		hook->private = &sc->inside;
	} else if (strcmp(name, NG_DRAWBRIDGE_HOOK_OUTSIDE) == 0) {
		sc->outside.hook = hook;
		hook->private = &sc->outside;
	} else if (strcmp(name, NG_DRAWBRIDGE_HOOK_MIRROR) == 0) {
		sc->mirror.hook = hook;
		hook->private = &sc->mirror;
	} else
		return (EINVAL);
	return (0);
}
 
/*
 * Receive a control message
 */ 
static int
ng_drawbridge_rcvmsg(node_p node, struct ng_mesg *msg, const char *retaddr,
	             struct ng_mesg **rptr)
{ 
	const sc_p sc = node->private;
	struct ng_mesg *resp = NULL;
	int error = 0;

	switch (msg->header.typecookie) {
	case NGM_DRAWBRIDGE_COOKIE:
		switch (msg->header.cmd) {
		case NGM_DRAWBRIDGE_GET_STATS:
		case NGM_DRAWBRIDGE_CLR_STATS:
		case NGM_DRAWBRIDGE_GET_CLR_STATS:
                    {			
                        if (msg->header.cmd != NGM_DRAWBRIDGE_CLR_STATS) {
                                NG_MKRESPONSE(resp, msg,sizeof(sc->stats), M_NOWAIT);
				if (resp == NULL) {
					error = ENOMEM;
					goto done;
				}				
				bcopy(&sc->stats,resp->data,sizeof(sc->stats));				
                        }
                        if (msg->header.cmd != NGM_DRAWBRIDGE_GET_STATS)
				bzero(&sc->stats,sizeof(sc->stats));										
                        break;
		    }
		case NGM_DRAWBRIDGE_SET_FLAGS:
		case NGM_DRAWBRIDGE_GET_FLAGS:
		    {			
		    	if (msg->header.cmd != NGM_DRAWBRIDGE_GET_FLAGS) {
				struct ng_drawbridge_flags *flags;					
				if (msg->header.arglen != sizeof(*flags)) {
					error = EINVAL;
					goto done;
				}
				flags = (struct ng_drawbridge_flags *)msg->data;								
				bcopy(flags,&sc->flags,sizeof(*flags));
			}
		     	if (msg->header.cmd != NGM_DRAWBRIDGE_SET_FLAGS) {			        
			    	NG_MKRESPONSE(resp,msg,sizeof(sc->flags),M_NOWAIT);
				if (resp == NULL) {
					error = ENOMEM;
					goto done;
				}
				bcopy(&sc->flags,resp->data,sizeof(sc->flags));
			}			
			break;
		    } 
		case NGM_DRAWBRIDGE_GET_ADDRESS_TBL:
		case NGM_DRAWBRIDGE_SET_ADDRESS_TBL:
		    {
		    	if (msg->header.cmd != NGM_DRAWBRIDGE_GET_ADDRESS_TBL) {
				addrTreeReq *addr;	
			    	if (msg->header.arglen != sizeof(addrTreeReq)) {
					error = EINVAL;
					goto done;
				}
				addr = (addrTreeReq *)msg->data;
				error = addrTreeRequest(WRITE,addr);
			}
			if (msg->header.cmd != NGM_DRAWBRIDGE_SET_ADDRESS_TBL) {
				addrTreeReq *addr;
				if (msg->header.arglen != sizeof(addrTreeReq)) {
					error = EINVAL;
					goto done;
				}
				addr = (addrTreeReq *)msg->data;
				error = addrTreeRequest(READ,addr);
				NG_MKRESPONSE(resp,msg,sizeof(addrTreeReq),M_NOWAIT);
				if (resp == NULL) {
					error = ENOMEM;
					goto done;
				}
				bcopy(addr,resp->data,sizeof(addrTreeReq));
			}
			break;
		    }
		case NGM_DRAWBRIDGE_GET_CLASS_TBL:
		case NGM_DRAWBRIDGE_SET_CLASS_TBL:
		    {
		    	if (msg->header.cmd != NGM_DRAWBRIDGE_GET_CLASS_TBL) {	
				portListReq *port;
			    	if (msg->header.arglen != sizeof(portListReq)) {
					error = EINVAL;
					goto done;
				}
				port = (portListReq *)msg->data;				
				error = portListRequest(WRITE,port,&new);
			}
			if (msg->header.cmd != NGM_DRAWBRIDGE_SET_CLASS_TBL) {
				portListReq *port;
				if (msg->header.arglen != sizeof(portListReq)) {
					error = EINVAL;
					goto done;
				}
				port = (portListReq *)msg->data;
				error = portListRequest(READ,port,&active);				
				NG_MKRESPONSE(resp,msg,sizeof(portListReq),M_NOWAIT);
				if (resp == NULL) {
					error = ENOMEM;
					goto done;
				}
				bcopy(port,resp->data,sizeof(portListReq));
			}
			break;
		    }
		case NGM_DRAWBRIDGE_GET_ACCEPT_TBL:
		case NGM_DRAWBRIDGE_SET_ACCEPT_TBL:
		    {
		    	if (msg->header.cmd != NGM_DRAWBRIDGE_GET_ACCEPT_TBL) {	
				rejaccTableReq *rejacc;
			    	if (msg->header.arglen != sizeof(rejaccTableReq)) {
					error = EINVAL;
					goto done;
				}
				rejacc = (rejaccTableReq *)msg->data;				
				error = rejaccTableRequest(WRITE,T_ACCTBL,rejacc,&new);
			}
			if (msg->header.cmd != NGM_DRAWBRIDGE_SET_ACCEPT_TBL) {
				rejaccTableReq *rejacc;
				if (msg->header.arglen != sizeof(rejaccTableReq)) {
					error = EINVAL;
					goto done;
				}
				rejacc = (rejaccTableReq *)msg->data;
				error = rejaccTableRequest(READ,T_ACCTBL,rejacc,&active);				
				NG_MKRESPONSE(resp,msg,sizeof(rejaccTableReq),M_NOWAIT);
				if (resp == NULL) {
					error = ENOMEM;
					goto done;
				}
				bcopy(rejacc,resp->data,sizeof(rejaccTableReq));
			}
			break;
		    }
		case NGM_DRAWBRIDGE_GET_REJECT_TBL:
		case NGM_DRAWBRIDGE_SET_REJECT_TBL:
		    {
		    	if (msg->header.cmd != NGM_DRAWBRIDGE_GET_REJECT_TBL) {	
				rejaccTableReq *rejacc;
			    	if (msg->header.arglen != sizeof(rejaccTableReq)) {
					error = EINVAL;
					goto done;
				}
				rejacc = (rejaccTableReq *)msg->data;				
				error = rejaccTableRequest(WRITE,T_REJTBL,rejacc,&new);
			}
			if (msg->header.cmd != NGM_DRAWBRIDGE_SET_REJECT_TBL) {
				rejaccTableReq *rejacc;
				if (msg->header.arglen != sizeof(rejaccTableReq)) {
					error = EINVAL;
					goto done;
				}
				rejacc = (rejaccTableReq *)msg->data;
				error = rejaccTableRequest(READ,T_REJTBL,rejacc,&active);				
				NG_MKRESPONSE(resp,msg,sizeof(rejaccTableReq),M_NOWAIT);
				if (resp == NULL) {
					error = ENOMEM;
					goto done;
				}
				bcopy(rejacc,resp->data,sizeof(rejaccTableReq));
			}
			break;
		    }
		case NGM_DRAWBRIDGE_GET_OVERRIDE_TBL:
		case NGM_DRAWBRIDGE_SET_OVERRIDE_TBL:
		    {
		    	if (msg->header.cmd != NGM_DRAWBRIDGE_GET_OVERRIDE_TBL) {	
				overrideTableReq *over;
			    	if (msg->header.arglen != sizeof(overrideTableReq)) {
					error = EINVAL;
					goto done;
				}
				over = (overrideTableReq *)msg->data;				
				error = overrideTableRequest(WRITE,over,&new);
			}
			if (msg->header.cmd != NGM_DRAWBRIDGE_SET_OVERRIDE_TBL) {
				overrideTableReq *over;
				if (msg->header.arglen != sizeof(overrideTableReq)) {
					error = EINVAL;
					goto done;
				}
				over = (overrideTableReq *)msg->data;
				error = overrideTableRequest(READ,over,&active);				
				NG_MKRESPONSE(resp,msg,sizeof(overrideTableReq),M_NOWAIT);
				if (resp == NULL) {
					error = ENOMEM;
					goto done;
				}
				bcopy(over,resp->data,sizeof(overrideTableReq));
			}
			break;
		    }
		case NGM_DRAWBRIDGE_GET_VERSION:
		    {		  
			NG_MKRESPONSE(resp,msg,sizeof(versionReq),M_NOWAIT);
			if (resp == NULL) {
				error = ENOMEM;
				goto done;
			}
			strncpy(((versionReq *)resp->data)->version,VERSION,32);
			break;
		    }
		case NGM_DRAWBRIDGE_GET_STATUS:
		    {
		    	NG_MKRESPONSE(resp,msg,sizeof(Status),M_NOWAIT);
			if (resp == NULL) {
				error = ENOMEM;
				goto done;
			}
			bcopy(&sc->DBstatus,resp->data,sizeof(Status));
			break;
		    }
		case NGM_DRAWBRIDGE_MANAGE_TABLES:
		    {
		    	manageTablesReq *tbls;
		    	if (msg->header.arglen != sizeof(manageTablesReq)) {
				error = EINVAL;
				goto done;
			}
			tbls = (manageTablesReq *)msg->data;			
			error = manageTablesRequest(tbls);
			break;
		    }		
		case NGM_DRAWBRIDGE_GET_BRIDGE_ENTRY:
		    {
		    	bridgeReq *bridge;
		    	if (msg->header.arglen != sizeof(bridgeReq)) {
				error = EINVAL;
				goto done;
			}
			bridge = (bridgeReq *)msg->data;			
			error = bridgeRequest(READ,bridge);
			NG_MKRESPONSE(resp,msg,sizeof(bridgeReq),M_NOWAIT);
			if (resp == NULL) {
				error = ENOMEM;
				goto done;
			}
			bcopy(bridge,resp->data,sizeof(bridgeReq));
			break;
		    }	
		case NGM_DRAWBRIDGE_CLR_BRIDGE_TBL:
		    {
		     	bzero(hashTable,sizeof(hashTable));
			sc->stats.bridgeEntries = 0;
			break;
		    }	
		default:
			error = EINVAL;
			break;
		}
		break;
	default:
		error = EINVAL;
		break;
	} 
	if (rptr)
		*rptr = resp;
	else if (resp)
		FREE(resp, M_NETGRAPH);

done:
	FREE(msg, M_NETGRAPH);
	return (error);
}

/*
 * Receive data on a hook
 *
 * If data comes in on the inside hook, send a copy out the mirror hook, 
 * and then send the original onwards out through the outside hook.
 * Do the opposite for data coming in from the outside hook.
 *
 */
static int
ng_drawbridge_rcvdata(hook_p hook, struct mbuf *m, meta_p meta)
{
	const sc_p sc = hook->node->private;	
	struct hookinfo *const hinfo = (struct hookinfo *) hook->private;
	struct hookinfo *dst,*src, *dup;	
	int error = 0;
	int db_action;
	register struct mbuf *r,*m2;
	meta_p dummy = NULL;	
	
	/* Which hook? */
	if (hinfo == &sc->outside) {	
		src = &sc->outside;
		dst = &sc->inside;
	} else if (hinfo == &sc->inside) {
		src = &sc->inside;
		dst = &sc->outside;
	} else if (hinfo == &sc->mirror) {
		NG_FREE_DATA(m,meta);
		return(0);	
	} else
		panic("%s: no hook!", __FUNCTION__);
	
	/* Construct a duplicate packet to send to the mirror */
	dup = &sc->mirror;
	m2 = m_dup(m, M_NOWAIT);
	if (m2 == NULL) {
		NG_FREE_DATA(m, meta);
		return (ENOBUFS);
	}
	NG_SEND_DATA(error,dup->hook,m2,dummy);
	
	/* Test the packet against our ruleset */
	db_action = drawbridge_test_pkt(sc,hinfo,m);      	
		
	switch(db_action) {
	
		case DB_FORWARD_PKT:
			/* Deliver packet out dst hook */			
			NG_SEND_DATA(error,dst->hook,m,meta);			
		break;  		
			
		case DB_DROP_AND_ANSWER_PKT:
			/* Create and send response packet out the src hook */ 
			MGETHDR(r,M_DONTWAIT,MT_DATA);		
			if(r == NULL) {
				NG_FREE_DATA(m,meta);
				break;
			}			
			drawbridge_response_pkt(r,m);	
			NG_FREE_DATA(m,meta);
			NG_SEND_DATA(error,src->hook,r,dummy);
		break;	
			
		case DB_DROP_PKT:
			/* Drop packet */			
			NG_FREE_DATA(m,meta);		
    		break;
	}		
	return (0);
}

/*
 * Shutdown processing
 *
 * This is tricky. If we have both an outside and a inside hook, then we
 * probably want to extricate ourselves and leave the two peers
 * still linked to each other. Otherwise we should just shut down as
 * a normal node would.
 *
 */
static int
ng_drawbridge_rmnode(node_p node)
{
	const sc_p privdata = node->private;

	node->flags |= NG_INVALID;
	if (privdata->outside.hook && privdata->inside.hook)
		ng_bypass(privdata->outside.hook, privdata->inside.hook);
	ng_cutlinks(node);
	ng_unname(node);
	node->private = NULL;
	ng_unref(privdata->node);
	FREE(privdata, M_NETGRAPH);
	return (0);
}

/*
 * Hook disconnection
 */
static int
ng_drawbridge_disconnect(hook_p hook)
{
	struct hookinfo *const hinfo = (struct hookinfo *) hook->private;

	KASSERT(hinfo != NULL, ("%s: null info", __FUNCTION__));
	hinfo->hook = NULL;
	if (hook->node->numhooks == 0)
		ng_rmnode(hook->node);
	return (0);
}


static void
freeTree(addrTreeNode *tbl, u_int level)
{
        int i;

        if (!tbl || (level > 3)) {
		printf("freeTree called with invalid parameters! "
			"(tbl:%p, level:%d)\n", tbl, level);
                return;
	}

	if (level == 3) {
		free((addrTreeLeaf *)tbl, M_DEVBUF);
		return;
	}

        for (i=0; i<256; i++) {
                if (!tbl->flg[i] && tbl->u[i].p) {
			if (level == 2)
				free(tbl->u[i].pd, M_DEVBUF);
			else
				freeTree(tbl->u[i].p, level + 1);
		}
	}

	free(tbl, M_DEVBUF);
}

/*
 * addr is expected to be in host (reverse) byte order
 * This function only works correctly on little endian CPUs
 */
static int
addToTree(addrTreeNode *tbl, in_addr_tw addr, u_int bits, u_int16_t class)
{
	in_addr_tw mask;
	u_int32_t i, byte, start, end, n, level;
	addrTreePtr ptr;	

        if (!tbl || (bits > 32))
                return -1;

	/* Find the rightmost non zero byte in the mask */
	mask.s_addr = 0xFFFFFFFFUL << ((bits < 32) ? 32 - bits : 0);
	for (byte=0; (byte < 3) && (mask.s_b[byte] == 0); byte++);

	/* Determine the range of addresses to be added to the tree */
	start = addr.s_b[byte] & mask.s_b[byte];
	end = start + (u_int8_t) ~mask.s_b[byte];

	for (i=start; i<=end; i++) {

		ptr = (addrTreePtr)tbl;
		addr.s_b[byte] = i;

		/* 'byte' specifies how deep in the tree to go */
		/* 'level' keeps track of how deep we are: 0, 1, 2, or 3 */
		for (n=byte, level=0; n < 4; n++, level++) {

			/* Use the next byte of the ip address as an index */
			/* into the array of pointers/classes of this node. */
			register u_int8_t b = addr.s_b[3 - level];

			/* If n equal 3, then this is the last byte of */
			/* the ip address to be put in the tree.  */
			if (n == 3) {
				if (level == 3) {
					/* Level 3 contains only classes */
					ptr.pd->c[b] = class;

				} else {
					/* Levels 0-2 can contain either a   */
					/* class or a pointer.  If a pointer */
					/* is stored here, then prune the    */
					/* tree from here up.                */
					if (!ptr.p->flg[b] && ptr.p->u[b].p)
						freeTree(ptr.p->u[b].p,level+1);
					ptr.p->u[b].p = NULL;

					/* Now save the class in this slot.  */
					ptr.p->u[b].c = class;
					ptr.p->flg[b] = 1;
				}

			/* If this is a pointer... */
			} else if (!ptr.p->flg[b] && ptr.p->u[b].p) {

				/* then move up the tree one level */
				ptr.p = ptr.p->u[b].p;

			/* Still more to go but we ran into a class   */
			/* index.  Add another level and move the     */
			/* class index to each slot in the new level. */
			} else {
				register u_int j;
				addrTreePtr next;

				/* If we are at level 2, then the */
				/*  next level is different.      */
				if (level == 2) {
					u_int tmp_c = ptr.p->u[b].c;

					if (!(next.pd = MALLOC_LEAF))
						return 1;
					for (j=0; j<256; j++) {
						next.pd->c[j] = tmp_c;
					}
				} else {
					void *tmp_p = ptr.p->u[b].p;
					u_int tmp_f = ptr.p->flg[b];

					if (!(next.p = MALLOC_NODE))
						return 1;
					for (j=0; j<256; j++) {
						next.p->u[j].p = tmp_p;
						next.p->flg[j] = tmp_f;
					}
				}
				/* Replace the class with the new pointer,  */
				/* set the flag to 0 to indicate a pointer, */
				/* and then move up the tree one level.     */ 
				ptr.p->u[b].p = next.p;
				ptr.p->flg[b] = 0;
				ptr.p = next.p;
			}
		}
	}
	return 0;
}

static void
freeAllPointers(pointerStruct *p)
{

	if (p->addrTree)	freeTree(p->addrTree, 0);
	if (p->t_dst_in)	free(p->t_dst_in, M_DEVBUF);
	if (p->t_dst_out)	free(p->t_dst_out, M_DEVBUF);
        if (p->t_src_in)	free(p->t_src_in, M_DEVBUF);
        if (p->t_src_out)	free(p->t_src_out, M_DEVBUF);
        if (p->u_dst_in)	free(p->u_dst_in, M_DEVBUF);
        if (p->u_dst_out)	free(p->u_dst_out, M_DEVBUF);
        if (p->u_src_in)	free(p->u_src_in, M_DEVBUF);
        if (p->u_src_out)	free(p->u_src_out, M_DEVBUF);
        if (p->i_typ_in)	free(p->i_typ_in, M_DEVBUF);
        if (p->i_typ_out)	free(p->i_typ_out, M_DEVBUF);
        if (p->t_rst_in)	free(p->t_rst_in, M_DEVBUF);
        if (p->t_rst_out)	free(p->t_rst_out, M_DEVBUF);
        if (p->acceptTable)	free(p->acceptTable, M_DEVBUF);
        if (p->acceptTree)	freeTree(p->acceptTree, 0);
        if (p->rejectTable)	free(p->rejectTable, M_DEVBUF);
        if (p->rejectTree)	freeTree(p->rejectTree, 0);
        if (p->overrideTable)	free(p->overrideTable, M_DEVBUF);
        if (p->overrideTree)	freeTree(p->overrideTree, 0);

	bzero(p, sizeof(pointerStruct));
}

static int
initTables(pointerStruct *p, u_int16_t classes)
{
	/* Must have at least one class for default */
	if (classes == 0)
		classes = 1;

	/* Free all allocated memory referenced by pointer struct */
	freeAllPointers(p);

	/* Build a new default config */
	p->addrTree  = MALLOC_NODE;
	p->t_dst_in  = MALLOC_CLASS(classes);
	p->t_dst_out = MALLOC_CLASS(classes);
	p->t_src_in  = MALLOC_CLASS(classes);
	p->t_src_out = MALLOC_CLASS(classes);
	p->u_dst_in  = MALLOC_CLASS(classes);
	p->u_dst_out = MALLOC_CLASS(classes);
	p->u_src_in  = MALLOC_CLASS(classes);
	p->u_src_out = MALLOC_CLASS(classes);
	p->i_typ_in  = MALLOC_CLASS(classes);
	p->i_typ_out = MALLOC_CLASS(classes);
	p->t_rst_in  = MALLOC_CLASS(1);
	p->t_rst_out = MALLOC_CLASS(1);
	if ((p->addrTree == NULL) ||
	    (p->t_dst_in  == NULL) || (p->t_dst_out == NULL) ||
	    (p->t_src_in  == NULL) || (p->t_src_out == NULL) ||
	    (p->u_dst_in  == NULL) || (p->u_dst_out == NULL) ||
	    (p->u_src_in  == NULL) || (p->u_src_out == NULL) ||
	    (p->i_typ_in  == NULL) || (p->i_typ_out == NULL) ||
	    (p->t_rst_in  == NULL) || (p->t_rst_out == NULL)) {
		freeAllPointers(p);
		return ERROR_NO_MEMORY;
	}
	p->num_classes = classes;
	bzero(p->addrTree,  sizeof(addrTreeNode));
	bzero(p->t_dst_in,  classes * LIST_TABLE_SIZE);
	bzero(p->t_dst_out, classes * LIST_TABLE_SIZE);
	bzero(p->t_src_in,  classes * LIST_TABLE_SIZE);
	bzero(p->t_src_out, classes * LIST_TABLE_SIZE);
	bzero(p->u_dst_in,  classes * LIST_TABLE_SIZE);
	bzero(p->u_dst_out, classes * LIST_TABLE_SIZE);
	bzero(p->u_src_in,  classes * LIST_TABLE_SIZE);
	bzero(p->u_src_out, classes * LIST_TABLE_SIZE);
	bzero(p->i_typ_in,  classes * LIST_TABLE_SIZE);
	bzero(p->i_typ_out, classes * LIST_TABLE_SIZE);
	bzero(p->t_rst_in,  LIST_TABLE_SIZE);
	bzero(p->t_rst_out, LIST_TABLE_SIZE);

	/* Fill in the class table defaults */
        p->t_dst_in[0].begin = 2;	/* TCP - two entries follow */
        p->t_dst_in[1].begin = 25;	/* TCP - Mail in            */
        p->t_dst_in[1].end = 25;
        p->t_dst_in[2].begin = 53;	/* TCP - Name service in */
        p->t_dst_in[2].end = 53;

        p->t_dst_out[0].begin = 1;	/* TCP - one entry follows */
        p->t_dst_out[1].begin = 0;	/* TCP - Everything out    */
        p->t_dst_out[1].end = 0xFFFF;

        p->t_src_in[0].begin = 1;	/* TCP - one entry follows       */
        p->t_src_in[1].begin = 20;	/* TCP - FTP data connections in */
        p->t_src_in[1].end = 20;

        p->u_dst_in[0].begin = 3;	/* UDP - three entries follow       */
        p->u_dst_in[1].begin = 0;	/* UDP - Block TFTP & Portmapper in */
        p->u_dst_in[1].end = 68;
        p->u_dst_in[2].begin = 70;
        p->u_dst_in[2].end = 110;
        p->u_dst_in[3].begin = 112;
        p->u_dst_in[3].end = 0xFFFF;

	p->u_dst_out[0].begin = 1;	/* UDP - one entry follows */
	p->u_dst_out[1].begin = 0;	/* UDP - Everything out    */
	p->u_dst_out[1].end = 0xFFFF;

	p->i_typ_in[0].begin = 1;	/* ICMP - one entry follows     */
	p->i_typ_in[1].begin = 0;	/* ICMP - Everything in         */
	p->i_typ_in[1].end = 0xFF;

	p->i_typ_out[0].begin = 1;	/* ICMP - one entry follows     */
	p->i_typ_out[1].begin = 0;	/* ICMP - Everything out        */
	p->i_typ_out[1].end = 0xFF;

	return NO_ERROR;
}

static int
copyClass(int rw, portListReq *u, PortListEntry *kptr)
{
	if (kptr == NULL)
		return ERROR_NET_NOT_FOUND; /* XXX */
	if (rw == WRITE)
		return copyin(u->ptr, kptr, u->bytes);
	if (rw == READ)
		return copyout(kptr, u->ptr, u->bytes);
	return ERROR_INVALID_OP;
}

static int
portListRequest(int rw, portListReq *u, pointerStruct *p)
{	
	if (u->class + ((u->bytes / LIST_TABLE_SIZE) +
		        (u->bytes % LIST_TABLE_SIZE) ? 1 : 0) > p->num_classes)
		return ERROR_OUT_OF_RANGE;

	switch (u->id) {
	case T_DST_IN:
		return copyClass(rw, u ,&(p->t_dst_in[u->class * LIST_SIZE]));
	case T_DST_OUT:
		return copyClass(rw, u ,&(p->t_dst_out[u->class * LIST_SIZE]));
	case T_SRC_IN:
		return copyClass(rw, u ,&(p->t_src_in[u->class * LIST_SIZE]));
	case T_SRC_OUT:
		return copyClass(rw, u ,&(p->t_src_out[u->class * LIST_SIZE]));
	case U_DST_IN:
		return copyClass(rw, u ,&(p->u_dst_in[u->class * LIST_SIZE]));
	case U_DST_OUT:
		return copyClass(rw, u ,&(p->u_dst_out[u->class * LIST_SIZE]));
	case U_SRC_IN:
		return copyClass(rw, u ,&(p->u_src_in[u->class * LIST_SIZE]));
	case U_SRC_OUT:
		return copyClass(rw, u ,&(p->u_src_out[u->class * LIST_SIZE]));
	case I_TYP_IN:
		return copyClass(rw, u ,&(p->i_typ_in[u->class * LIST_SIZE]));
	case I_TYP_OUT:
		return copyClass(rw, u ,&(p->i_typ_out[u->class * LIST_SIZE]));
	case T_RST_IN:
		if (u->class > 0)
			return ERROR_OUT_OF_RANGE;
		return copyClass(rw, u, p->t_rst_in);
	case T_RST_OUT:
		if (u->class > 0)
			return ERROR_OUT_OF_RANGE;
		return copyClass(rw, u, p->t_rst_out);
	default:
		return ERROR_UNKNOWN_CLASS;
	}
}

/*
 * 'addr' is expected to be in host (reverse) byte order.
 * This function only works correctly on little endian CPUs.
 */
static int
treeLookup(addrTreeNode *tbl, u_int32_t addr, u_int16_t max_class)
{
        register addrTreePtr t1,t2;
	register in_addr_tw ip;
	
	if (tbl == NULL)
		return DEFAULT_ACCESS;
	ip = (in_addr_tw)addr;

        /*
         * lookup first octet of address - flg=1: class flg=0: pointer
         */
        if ((t1.p = tbl->u[ip.s_b[3]].p) == NULL)
                return DEFAULT_ACCESS;
        if (tbl->flg[ip.s_b[3]])
                return ((t1.c < max_class) ? t1.c : DEFAULT_ACCESS);

        /*
         * lookup second octet of address - flg=1: class flg=0: pointer
         * t1.p = tbl->u[ip.s_b[3]].p
         */
        if ((t2.p = t1.p->u[ip.s_b[2]].p) == NULL)
                return DEFAULT_ACCESS;
        if (t1.p->flg[ip.s_b[2]])
                return ((t2.c < max_class) ? t2.c : DEFAULT_ACCESS);

        /*
         * lookup third octet of address - flg=1: class flg=0: pointer
         * t2.p = tbl->u[ip.s_b[3]].p->u[ip.s_b[2]].p
         */
        if ((t1.pd = t2.p->u[ip.s_b[1]].pd) == NULL)
                return DEFAULT_ACCESS;
        if (t2.p->flg[ip.s_b[1]])
                return ((t1.c < max_class) ? t1.c : DEFAULT_ACCESS);

        /*
         * lookup fourth octet of address - always a class
         * t1.pd = tbl->u[ip.s_b[3]].p->u[ip.s_b[2]].p->u[ip.s_b[1]].pd
         */
        t2.c = t1.pd->c[ip.s_b[0]];
        return ((t2.c < max_class) ? t2.c : DEFAULT_ACCESS);
}

static int
addrTreeRequest(int rw, addrTreeReq *u)
{
	register u_int i;
	
	switch (rw) {
        case READ:
		/* Lookup the specified IP cidr(s) in the 'active' config */
		for (i=0; (i < 128) && (i < u->num); i++)
			u->class[i] = treeLookup(active.addrTree,
					u->addr[i].s_addr, active.num_classes);
		break;
	
        case WRITE:
		if (new.addrTree == NULL) 
			return ERROR_NET_NOT_FOUND; /* XXX */

		/* Add the specified IP cidr(s) to the 'new' config */
		for (i=0; (i < u->num) && (i < 128); i++) {
			if (addToTree(new.addrTree, u->addr[i],
				     u->bits[i], u->class[i])) {
				freeTree(new.addrTree, 0);
				new.addrTree = NULL;
				return ERROR_NO_MEMORY;
			}
		}
		break;

	default:
		return ERROR_INVALID_OP;
	}

	return NO_ERROR;
}

static int
rejaccTableRequest(int rw, int id, rejaccTableReq *u, pointerStruct *p)
{
	register u_int i;
	RejAccTableEntry *table;
	addrTreeNode *tree;
	
	if ((id != T_REJTBL) && (id != T_ACCTBL))
		return ERROR_INVALID_OP;

	switch (rw) {
        case READ:
		u->bytes = (id == T_REJTBL) ? p->reject_bytes : p->accept_bytes;
		u->flag = treeLookup((id == T_REJTBL) ? active.rejectTree :
					active.acceptTree, u->addr, 2);
		table = (id == T_REJTBL) ? p->rejectTable : p->acceptTable;
		if ((u->bytes != 0) && (table == NULL))
			return ERROR_NET_NOT_FOUND; /* XXX */
		/* If it exists and the user requested it, */
		/* copy the table up to user space.        */
		if ((u->ptr != NULL) && (table != NULL)) {
			if ((i = copyout(table, u->ptr, u->bytes)) != 0)
				return i;
		}
		break;
	
        case WRITE:
		/* Perform a few sanity checks */
		if ((u->bytes == 0) || (u->bytes % sizeof(RejAccTableEntry)))
			return ERROR_INVALID_OP;
		if (id == T_REJTBL) {
			if ((p->rejectTable != NULL) || (p->rejectTree != NULL))
				return ERROR_NET_NOT_FOUND; /* XXX */
		} else {
			if ((p->acceptTable != NULL) || (p->acceptTree != NULL))
				return ERROR_NET_NOT_FOUND; /* XXX */
		}

		/*
		 * Allocate the memory for the table and tree.  The tree
		 * is used for packet filtering.  The table is only used
		 * to keep track of the networks/hosts to show to the user.
		 */
		table = (RejAccTableEntry *)malloc(u->bytes,M_DEVBUF,M_NOWAIT);
		if (table == NULL)
			return ERROR_NO_MEMORY;
		tree = MALLOC_NODE;
		if (tree == NULL) {
			free(table, M_DEVBUF);
			return ERROR_NO_MEMORY;
		}
		bzero(tree, sizeof(addrTreeNode));
		bzero(table, u->bytes);

		/* Copy the table in from user space */
		if ((i = copyin(u->ptr, table, u->bytes)) != 0)
			return i;
		/*
		 * Populate the tree from the table - keep the table around
		 * to spit back out when requested by the user interface.
		 */
		for (i=u->bytes/sizeof(RejAccTableEntry); i; i--) {
			register int r;

			if ((r=addToTree(tree, table[i-1].network,
				      table[i-1].bits, table[i-1].flag))) {
				freeTree(tree, 0);
				free(table, M_DEVBUF);
				return (r>0) ? ERROR_NO_MEMORY :
					ERROR_INVALID_OP;
			}
		}
		if (id == T_REJTBL) {
			p->rejectTable = table;
			p->rejectTree = tree;
			p->reject_bytes = u->bytes;
		} else {
			p->acceptTable = table;
			p->acceptTree = tree;
			p->accept_bytes = u->bytes;
		}
		break;

	default:
		return ERROR_INVALID_OP;
	}

	return NO_ERROR;
}

static int
overrideTableRequest(int rw, overrideTableReq *u, pointerStruct *p)
{
	u_int i,j,r;
	OverrideTableEntry *table;
	addrTreeNode *tree;
	
	switch (rw) {
        case READ:
		u->bytes = p->override_bytes;
		u->index = treeLookup(active.overrideTree,
				      u->addr, active.num_override);
		table = p->overrideTable;
		if ((u->bytes != 0) && (table == NULL))
			return ERROR_NET_NOT_FOUND; /* XXX */
		/* If it exists and the user requested it, */
		/* copy the table up to user space.        */
		if ((u->ptr != NULL) && (table != NULL)) {
			if ((i = copyout(table, u->ptr, u->bytes)) != 0)
				return i;
		}
		break;
	
        case WRITE:
		/* Perform a few sanity checks */
		if ((u->bytes == 0) || (u->bytes % sizeof(OverrideTableEntry)))
			return ERROR_INVALID_OP;
		if ((p->overrideTable != NULL) || (p->overrideTree != NULL))
			return ERROR_NET_NOT_FOUND; /* XXX */

		/* Allocate the memory for the table and tree. */
		table = (OverrideTableEntry *)
				malloc(u->bytes, M_DEVBUF, M_NOWAIT);
		if (table == NULL)
			return ERROR_NO_MEMORY;
		tree = MALLOC_NODE;
		if (tree == NULL) {
			free(table, M_DEVBUF);
			return ERROR_NO_MEMORY;
		}
		bzero(tree, sizeof(addrTreeNode));
		bzero(table, u->bytes);

		/* Copy the table in from user space */
		if ((i = copyin(u->ptr, table, u->bytes)) != 0)
			return i;

		/* Populate the tree from the table */
		for (i=u->bytes/sizeof(OverrideTableEntry),j=1; i; i--,j++) {
			if ((r=addToTree(tree, table[i-1].network,
				      table[i-1].bits, j))) {
				freeTree(tree, 0);
				free(table, M_DEVBUF);
				return (r>0) ? ERROR_NO_MEMORY :
					ERROR_INVALID_OP;
			}
		}
		p->overrideTable = table;
		p->overrideTree = tree;
		p->override_bytes = u->bytes;
		p->num_override = (u->bytes / sizeof(OverrideTableEntry)) + 1;
		break;

	default:
		return ERROR_INVALID_OP;
	}

	return NO_ERROR;
}

static int
manageTablesRequest(manageTablesReq *u)
{
	pointerStruct tmp;
	int rcode;
	
	switch (u->command) {

	/* Move the 'new' config to 'active' and deallocate the old config */
	case ACTIVATE_NEW:
		if ((new.num_classes == 0) || (new.addrTree == NULL) ||
		    (new.t_dst_in == NULL) || (new.t_dst_out == NULL) ||
		    (new.t_src_in == NULL) || (new.t_src_out == NULL) ||
		    (new.u_dst_in == NULL) || (new.u_dst_out == NULL) ||
		    (new.u_src_in == NULL) || (new.u_src_out == NULL) ||
		    (new.i_typ_in == NULL) || (new.i_typ_out == NULL) ||
		    (new.t_rst_in == NULL) || (new.t_rst_out == NULL))
			return ERROR_NET_NOT_FOUND; /* XXX */
		memcpy(&tmp, &active, sizeof(pointerStruct));
		memcpy(&active, &new, sizeof(pointerStruct));
		bzero(&new, sizeof(pointerStruct));
		freeAllPointers(&tmp);
		break;

	/* Remove all custom filters and revert to defaults */
	case CLEAR_ALL:
		/* (Re)initialize the 'new' config. */
		rcode = initTables(&new, 1);
		if (rcode)
			return rcode;
		memcpy(&tmp, &active, sizeof(pointerStruct));
		/* Move the 'new' config to the 'active' config */
		memcpy(&active, &new, sizeof(pointerStruct));
		bzero(&new, sizeof(pointerStruct));
		/* Deallocate the old config */
		freeAllPointers(&tmp);
		break;

	/* (Re)initialize the 'new' config.  Needs to   */
	/* be done before loading a new set of filters. */
	case INIT_NEW:
		rcode = initTables(&new, u->val);
		if (rcode)
			return rcode;
		break;

	}

	return NO_ERROR;
}

static int
bridgeRequest(int rw, bridgeReq *r)
{
	
	if (rw != READ)
		return ERROR_INVALID_OP;

	if (r->index >= MAX_BRIDGE_ADDRESSES)
		return ERROR_OUT_OF_RANGE;

	bcopy(&hashTable[r->index], &r->t, sizeof(r->t));

	return NO_ERROR;
}

/*
 * Parameters are expected to be in host byte order
 */
static int
checkIncomingTcp(u_int32_t srcAddr, u_int32_t dstAddr,
		 u_int16_t srcPort, u_int16_t dstPort,const sc_p sc)
{
	register int i;
	int class;
	PortListEntry *portList;
	
	/*
	 * Check for multicast
	 */
	if (IN_CLASSD(dstAddr)) {
		if (sc->flags.discardMulticast) {
			sc->stats.f_outsideTcpMcast++;	
			syslogMessage(SYSL_IN_CLASSD, IPPROTO_TCP,
			      	srcAddr, dstAddr, (int)srcPort, (int)dstPort);		
			return DISCARD_PKT;
		}
		return ALLOW_PKT;
	}

	/*
	 * Do the lookup to get the class.
	 */
	class = treeLookup(active.addrTree, dstAddr, active.num_classes);
	portList = &active.t_dst_in[class * LIST_SIZE];

	/*
	 * See if the destination port is allowed by the 'in' port list
	 */
	for (i=1; i <= portList[0].begin; i++) {
		if ((dstPort >= portList[i].begin) &&
		    (dstPort <= portList[i].end))
	    		return ALLOW_PKT;
	} 	

	/*
	 * If the destination port is not allowed but is over 900, then
	 * check the source port.  NOTE: This is 900 instead of 1024 because
	 * some broken FTP implementations use ports below 1024.  If you want
	 * to enforce the use of non-privileged ports, change this to >= 1024.
	 */
	if (dstPort > 1024) {
		portList = &active.t_src_in[class * LIST_SIZE];

		for (i=1; i <= portList[0].begin; i++) {
			if ((srcPort >= portList[i].begin) &&
			    (srcPort <= portList[i].end))
				return ALLOW_PKT;
		}
	}

	sc->stats.f_outsideTcpPort++;
	syslogMessage(SYSL_IN_PORT, IPPROTO_TCP,
		      srcAddr, dstAddr, (int)srcPort, (int)dstPort);	
	/* Check the TCP reset table */	
	for (i=1; i <= active.t_rst_in[0].begin; i++) {
		if ((dstPort >= active.t_rst_in[i].begin) &&
		    (dstPort <= active.t_rst_in[i].end)) {
			sc->stats.outsideTCPreset++;
			return DISCARD_AND_ANSWER_PKT;
		}
	}
	
	return DISCARD_PKT;
}

/*
 * Parameters are expected to be in host byte order
 */
static int
checkOutgoingTcp(u_int32_t srcAddr, u_int32_t dstAddr,
		 u_int16_t srcPort, u_int16_t dstPort,const sc_p sc)
{
	register int i, j;
	int class;
	PortListEntry *portList;
	
	/*
	 * Check for multicast
	 */
	if (IN_CLASSD(dstAddr)) {
		if (sc->flags.discardMulticast) {
			sc->stats.f_insideTcpMcast++;
			syslogMessage(SYSL_OUT_CLASSD,IPPROTO_TCP,
				srcAddr, dstAddr, (int)srcPort, (int)dstPort);			
			return DISCARD_PKT;
		}
		return ALLOW_PKT;
	}

	/*
	 * Do the lookup to get the class.
	 */
	class = treeLookup(active.addrTree, srcAddr, active.num_classes);
	portList = &active.t_dst_out[class * LIST_SIZE];

	/*
	 * See if the destination port is allowed by the 'out' port list
	 */
	for (i=1; i <= portList[0].begin; i++) {
		if ((dstPort >= portList[i].begin) &&
		    (dstPort <= portList[i].end))
			return ALLOW_PKT;
	}

	/*
	 * If the destination port is not allowed but is over 900, then
	 * check the source port.  NOTE: This is 900 instead of 1024 because
	 * some broken FTP implementations use ports below 1024.  If you want
	 * to enforce the use of non-privileged ports, change this to >= 1024.
	 */
	if (dstPort > 1024) {
		portList = &active.t_src_out[class * LIST_SIZE];

		for (i=1; i <= portList[0].begin; i++) {
			if ((srcPort >= portList[i].begin) &&
			    (srcPort <= portList[i].end))
				return ALLOW_PKT;
		}
	}

	/*
	 * Still not allowed so check the outgoing packet to see
	 * if the destination is in the override table allow list.
	 */
	if ((i = treeLookup(active.overrideTree,
			    dstAddr, active.num_override)) == 0) {
		sc->stats.f_insideTcpPort++;
		syslogMessage(SYSL_OUT_PORT, IPPROTO_TCP,
			srcAddr, dstAddr, (int)srcPort, (int)dstPort);		
		/* Check the TCP reset table */
		for (i=1; i <= active.t_rst_out[0].begin; i++) {
			if ((dstPort >= active.t_rst_out[i].begin) &&
			    (dstPort <= active.t_rst_out[i].end)) {
				sc->stats.insideTCPreset++;
				return DISCARD_AND_ANSWER_PKT;
			}
		}
    		/* Dest addr was not in the override table so discard pkt */
		return DISCARD_PKT;
	}

	/*
	 * It's in the override table so check the destination port.  The
	 * override table index is offset by one so decrement i by one first.
	 */
	for (i--,j=1; j <= active.overrideTable[i].port[0].begin; j++) {
		if ((dstPort >= active.overrideTable[i].port[j].begin) &&
		    (dstPort <= active.overrideTable[i].port[j].end))
	    		return ALLOW_PKT;
	}

	/*
	 * If we made it this far, the destination port is not allowed
	 */
	sc->stats.f_insideTcpPortOverride++;	
	syslogMessage(SYSL_OUT_OVERRIDE, IPPROTO_TCP,
		      srcAddr, dstAddr, (int)srcPort, (int)dstPort);
		      
	/* Check the TCP reset table */
	for (i=1; i <= active.t_rst_out[0].begin; i++) {
		if ((dstPort >= active.t_rst_out[i].begin) &&
		    (dstPort <= active.t_rst_out[i].end)) {
			sc->stats.insideTCPreset++;
			return DISCARD_AND_ANSWER_PKT;
		}
	}
	return DISCARD_PKT;
}
	
/*
 * Parameters are expected to be in host byte order
 */
static int
checkIncomingUdp(u_int32_t srcAddr, u_int32_t dstAddr,
		 u_int16_t srcPort, u_int16_t dstPort,const sc_p sc)
{
	register int i;
	int class;
	PortListEntry *portList;
	
	/*
	 * Check for multicast
	 */
	if (IN_CLASSD(dstAddr)) {
		if (sc->flags.discardMulticast) {
			sc->stats.f_outsideUdpMcast++;	
			syslogMessage(SYSL_IN_CLASSD,IPPROTO_UDP,
				srcAddr, dstAddr, (int)srcPort, (int)dstPort);		
			return DISCARD_PKT;
		}
		return ALLOW_PKT;
	}

	/*
	 * Do the lookup to get the class.
	 */
	class = treeLookup(active.addrTree, dstAddr, active.num_classes);

	/*
	 * See if the destination port is allowed by the 'udp' port list.
	 */
	portList = &active.u_dst_in[class * LIST_SIZE];
	for (i=1; i <= portList[0].begin; i++) {
		if ((dstPort >= portList[i].begin) &&
		    (dstPort <= portList[i].end))
			return ALLOW_PKT;
	}

        /*
         * If the destination port is not allowed then check the source port.
         */
	if (dstPort >= 1024) {
		portList = &active.u_src_in[class * LIST_SIZE];
		for (i=1; i <= portList[0].begin; i++) {
			if ((srcPort >= portList[i].begin) &&
			    (srcPort <= portList[i].end))
				return ALLOW_PKT;
		}
	}

	/*
	 * If we made it this far, the destination port is not allowed
	 */
	sc->stats.f_outsideUdpPort++;
	syslogMessage(SYSL_IN_PORT, IPPROTO_UDP,
		      srcAddr, dstAddr, (int)srcPort, (int)dstPort);	
	return DISCARD_PKT;
}


/*
 * Parameters are expected to be in host byte order
 */
static int
checkOutgoingUdp(u_int32_t srcAddr, u_int32_t dstAddr,
		 u_int16_t srcPort, u_int16_t dstPort,const sc_p sc)
{
	register int i;
	int class;
	PortListEntry *portList;
	
	/*
	 * Check for multicast
	 */
	if (IN_CLASSD(dstAddr)) {
		if (sc->flags.discardMulticast) {
			sc->stats.f_insideUdpMcast++;	
			syslogMessage(SYSL_OUT_CLASSD,IPPROTO_UDP,
				srcAddr, dstAddr, (int)srcPort, (int)dstPort);		
			return DISCARD_PKT;
		}
		return ALLOW_PKT;
	}

	/*
	 * Do the lookup to get the class.
	 */
	class = treeLookup(active.addrTree, srcAddr, active.num_classes);

	/*
	 * See if the destination port is allowed by the 'udp' port list.
	 */
	portList = &active.u_dst_out[class * LIST_SIZE];
	for (i=1; i <= portList[0].begin; i++) {
		if ((dstPort >= portList[i].begin) &&
		    (dstPort <= portList[i].end))
			return ALLOW_PKT;
	}

        /*
         * If the destination port is not allowed then check the source port.
         */
	if (dstPort >= 1024) {
		portList = &active.u_src_out[class * LIST_SIZE];
		for (i=1; i <= portList[0].begin; i++) {
			if ((srcPort >= portList[i].begin) &&
			    (srcPort <= portList[i].end))
				return ALLOW_PKT;
		}
	}

	/*
	 * If we made it this far, the destination port is not allowed
	 */
	sc->stats.f_insideUdpPort++;	
	syslogMessage(SYSL_OUT_PORT, IPPROTO_UDP,
		      srcAddr, dstAddr, (int)srcPort, (int)dstPort);
		      
	return DISCARD_PKT;
}

/*
 * Parameters are expected to be in host byte order
 */
static int
checkIncomingIcmp(u_int32_t srcAddr, u_int32_t dstAddr, u_int8_t type,const sc_p sc)
{
	register int i;
	PortListEntry *portList;
	
	portList = &active.i_typ_in[ treeLookup(active.addrTree,
				dstAddr, active.num_classes) * LIST_SIZE ];

	/*
	 * See if the ICMP type is allowed by the 'icmp' port list.
	 */
	for (i=1; i <= portList[0].begin; i++) {
		if ((type >= portList[i].begin) &&
		    (type <= portList[i].end))
			return ALLOW_PKT;
	}

	/*
	 * If we made it here, the ICMP type is not allowed
	 */
	sc->stats.f_outsideIcmpType++;	
	syslogMessage(SYSL_IN_TYPE, srcAddr, dstAddr, (int)type-1);
	return DISCARD_PKT;
}

/*
 * Parameters are expected to be in host byte order
 */
static int
checkOutgoingIcmp(u_int32_t srcAddr, u_int32_t dstAddr, u_int8_t type,const sc_p sc)
{
	register int i;
	PortListEntry *portList;
	
	portList = &active.i_typ_out[ treeLookup(active.addrTree,
				srcAddr, active.num_classes) * LIST_SIZE ];

	/*
	 * See if the ICMP type is allowed by the 'icmp' port list.
	 */
	for (i=1; i <= portList[0].begin; i++) {
		if ((type >= portList[i].begin) &&
		    (type <= portList[i].end))
			return ALLOW_PKT;
	}

	/*
	 * If we made it here, the ICMP type is not allowed
	 */
	sc->stats.f_insideIcmpType++;	
	syslogMessage(SYSL_OUT_TYPE, srcAddr, dstAddr, (int)type-1);
	return DISCARD_PKT;
}

static int
checkRejectTable(struct ip *ipHeader, u_int32_t srcAddr, u_int32_t dstAddr,const sc_p sc)
{

	/* If the reject table is empty, then forward all packets */
	if (active.rejectTree == NULL)
		return ALLOW_PKT;

	/* If srcAddr is in the tree and it's class is 1, then discard pkt */
	if (treeLookup(active.rejectTree, srcAddr, 2) == 1) {
		sc->stats.f_outsideRejectTable++;	
		syslogMessage(SYSL_IN_REJECT, (int)ipHeader->ip_p,
			srcAddr, dstAddr);	
		return DISCARD_PKT;
	}

	/* Process by default */
	return ALLOW_PKT;
}

static int
checkAcceptTable(struct ip *ipHeader, u_int32_t srcAddr, u_int32_t dstAddr,const sc_p sc)
{
	/* If the accept table is empty, then forward all packets */
	if (active.acceptTree == NULL)
		return ALLOW_PKT;

	/* If srcAddr is in the tree and it's class is 1, then process pkt */
	if (treeLookup(active.acceptTree, srcAddr, 2) == 1)
		return ALLOW_PKT;

	/* Discard by default */
	sc->stats.f_insideAcceptTable++;
	syslogMessage(SYSL_OUT_ACCEPT, (int)ipHeader->ip_p,
			srcAddr, dstAddr);	
	return DISCARD_PKT;
}


/*
 * Parameters 'protocol' and 'protolen' are expected to be in host byte order
 * The packet pointed to by 'pkt' is, of course, in network btye order
 */
static int
checkIncomingPacket(u_int16_t protocol, u_int8_t *pkt,
		    u_int protolen, int checkAccept,const sc_p sc)
{
	struct ip *ipHeader;
	struct tcphdr *tcpHeader;
	struct udphdr *udpHeader;
	struct icmp *icmpHeader;
	u_int32_t srcAddr, dstAddr;
	u_int16_t ip_len;
	
	switch (protocol) {

	case ETHERTYPE_IP:               
		ipHeader = (struct ip *)pkt;
		srcAddr = ntohl(ipHeader->ip_src.s_addr);
		dstAddr = ntohl(ipHeader->ip_dst.s_addr);

		if (checkAccept) {
			if (checkAcceptTable(ipHeader, srcAddr, dstAddr, sc)
			    == DISCARD_PKT)
				return DISCARD_PKT;
		} else {
			if (checkRejectTable(ipHeader, srcAddr, dstAddr, sc)
			    == DISCARD_PKT)
				return DISCARD_PKT;
		}

		/*
		 * Pass all IP fragments that do not have offset 0 (beginning
		 * of the packet) without checking since the TCP/UDP
		 * headers are not in this packet.
		 */
		if ((ipHeader->ip_off & SW_IP_OFFMASK) != 0) {
			/*
			 * If option is set, then discard pkts with offset of 1
			 */
			if (sc->flags.discardSuspectOffset &&
			    (ntohs(ipHeader->ip_off & SW_IP_OFFMASK) == 1) &&
			    (ipHeader->ip_p == IPPROTO_TCP)) {
				sc->stats.f_outsideSuspectOffset++;	
				syslogMessage(SYSL_IN_OFFSET, IPPROTO_TCP,
					srcAddr, dstAddr);	
				return DISCARD_PKT;
			}

			/*
			 * If option is set, then discard fragmented ICMP pkts
			 */
			if (sc->flags.discardFragmentedICMP &&
			    (ipHeader->ip_p == IPPROTO_ICMP)) {

				sc->stats.f_outsideIcmpFrag++;
				syslogMessage(SYSL_IN_FRAG, IPPROTO_ICMP,
					srcAddr, dstAddr);			    	
				return DISCARD_PKT;
			}

			return ALLOW_PKT;
		}
		switch (ipHeader->ip_p) {
		
		case IPPROTO_TCP:			
			tcpHeader = (struct tcphdr *)(pkt + 
						(ipHeader->ip_hl << 2)); 
			/*
			 * Make sure this packet (fragment) includes enough of
			 * the TCP header before making the other checks.
			 * Otherwise a SYN can sneak through since we might be
			 * trying to test something which is actually in the
			 * next fragment.
			 *
			 * NOTE: we drop all of these since we do not keep any
			 *       state between fragments.
			 *
			 * Many thanks to Uwe Ellermann at DFN-CERT for
			 * reporting this problem.
			 */
			ip_len = ntohs(ipHeader->ip_len);
			if ((ip_len > protolen) ||
			    ((ip_len - (ipHeader->ip_hl << 2)) < 14)) {
				sc->stats.f_outsideShortTcpHdr++;
				syslogMessage(SYSL_IN_LENGTH, IPPROTO_TCP,
					srcAddr, dstAddr,
					ntohs(tcpHeader->th_sport),
					ntohs(tcpHeader->th_dport));	
				return DISCARD_PKT;
			}

			/* Check for "ACKless SYN" */
			if ((tcpHeader->th_flags & (TH_SYN|TH_ACK)) == TH_SYN) {
				return checkIncomingTcp(srcAddr, dstAddr,
						ntohs(tcpHeader->th_sport),
						ntohs(tcpHeader->th_dport), sc);
			}
			break;

		case IPPROTO_UDP:  			
			udpHeader = (struct udphdr *)(pkt + 
						(ipHeader->ip_hl << 2));
			/*
			 * Make sure this packet (fragment) includes enough
			 * of the UDP header before making the other checks.
			 */
			ip_len = ntohs(ipHeader->ip_len);
			if ((ip_len > protolen) ||
			    ((ip_len - (ipHeader->ip_hl << 2)) < 4)) {			    
				sc->stats.f_outsideShortUdpHdr++;
				syslogMessage(SYSL_IN_LENGTH, IPPROTO_UDP,
					srcAddr, dstAddr,
					ntohs(udpHeader->uh_sport),
					ntohs(udpHeader->uh_dport));				
				return DISCARD_PKT;
			}

			/* Check UDP protocols. */

			return checkIncomingUdp(srcAddr, dstAddr,
						ntohs(udpHeader->uh_sport),
						ntohs(udpHeader->uh_dport), sc);
			break;

		case IPPROTO_ICMP:			
			icmpHeader = (struct icmp *)
				     (pkt + (ipHeader->ip_hl << 2));

			/*
			 * If option is set, then discard fragmented ICMP pkts
			 */
			if (sc->flags.discardFragmentedICMP &&
			    (ipHeader->ip_off & IP_MF) != 0) {
				sc->stats.f_outsideIcmpFrag++;	
				syslogMessage(SYSL_IN_FRAG, IPPROTO_ICMP,
					srcAddr, dstAddr);		    	
				return DISCARD_PKT;
			}

			/*
			 * If option is set, discard attack ICMP pkts.
			 *
			 * If the ICMP type, code, identifier, and seq num all
			 * equal 0, this may be an ICMP echo reply attack pkt.
			 */
			if (sc->flags.discardAttackICMP &&
			    (*((u_int16_t *)icmpHeader) == 0) && /*type & code*/
			    (icmpHeader->icmp_void == 0)) {      /*id & seq*/

				if (icmpHeader->icmp_cksum == 0xFFFF) {
				
					/* This is a "smurf" attack */
					sc->stats.f_outsideSmurfDos++;	
					syslogMessage(SYSL_IN_DOS,
						      srcAddr, dstAddr,
						      DOS_SMURF);		    		
					return DISCARD_PKT;

				} else if (*((u_int16_t *)icmpHeader->icmp_data)
					   == htons(0x4500)) {
					/* This is a "pong" attack */
					sc->stats.f_outsidePongDos++;	
					syslogMessage(SYSL_IN_DOS,
						      srcAddr, dstAddr,
						      DOS_PONG);		    		
					return DISCARD_PKT;
				}
			}

			return checkIncomingIcmp(srcAddr, dstAddr,
						 icmpHeader->icmp_type,sc);			
			break;

		default:
			if (sc->flags.discardOtherIp) {

				sc->stats.f_outsideOtherIp++;	
				syslogMessage(SYSL_IN_PROT, (int)ipHeader->ip_p,
					srcAddr, dstAddr);			
				return DISCARD_PKT;
			}

			/* Everything is allowed so far. */
			break;
		}

		break;

	case ETHERTYPE_ARP:
	case ETHERTYPE_REVARP:  		
		/* pass these guys through no matter what */
		break;

	default:
		/* Only pass the rest if told to do so. */
		if (sc->flags.discardNonIp) {			
			sc->stats.f_outsideNonIp++;	
			syslogMessage(SYSL_IN_FILTER, protocol);		
			return DISCARD_PKT;
		}
	}

	return ALLOW_PKT;
}


/*
 * Parameters 'protocol' and 'protolen' are expected to be in host byte order
 * The packet pointed to by 'pkt' is, of course, in network btye order
 */
static int
checkOutgoingPacket(u_int16_t protocol, u_int8_t *pkt, u_int protolen,const sc_p sc)
{
	struct ip *ipHeader;
	struct tcphdr *tcpHeader;
	struct udphdr *udpHeader;
	struct icmp *icmpHeader;
	u_int16_t ip_len;
	u_int32_t srcAddr, dstAddr;

	switch (protocol) {

	case ETHERTYPE_IP:
		ipHeader = (struct ip *)pkt;
		srcAddr = ntohl(ipHeader->ip_src.s_addr);
		dstAddr = ntohl(ipHeader->ip_dst.s_addr);

		if (checkAcceptTable(ipHeader,srcAddr,dstAddr,sc) == DISCARD_PKT)
			return DISCARD_PKT;

		/*
		 * Pass all IP fragments that do not have offset 0 (beginning
		 * of the packet) without checking since the TCP/UDP
		 * headers are not in this packet.  An area for improvement
		 * would be to cache session info so we could drop all
		 * disallowed fragments also instead of just the first one.
		 */
		if ((ipHeader->ip_off & SW_IP_OFFMASK) != 0) {

			if (sc->flags.discardSuspectOffset &&
			    (ntohs(ipHeader->ip_off & SW_IP_OFFMASK) == 1) &&
			    (ipHeader->ip_p == IPPROTO_TCP)) {

				sc->stats.f_insideSuspectOffset++;	
				syslogMessage(SYSL_OUT_OFFSET, IPPROTO_TCP,
					srcAddr, dstAddr);		
				return DISCARD_PKT;
			}

			/*
			 * If option is set, then discard fragmented ICMP pkts
			 */
			if (sc->flags.discardFragmentedICMP &&
			    (ipHeader->ip_p == IPPROTO_ICMP)) {

				sc->stats.f_insideIcmpFrag++;	
				syslogMessage(SYSL_OUT_FRAG, IPPROTO_ICMP,
					srcAddr, dstAddr);		    	
				return DISCARD_PKT;
			}

			return ALLOW_PKT;
		}

		switch (ipHeader->ip_p) {

		case IPPROTO_TCP:
			tcpHeader = (struct tcphdr *)(pkt + 
						(ipHeader->ip_hl << 2)); 
			/*
			 * Make sure this packet (fragment) includes enough
			 * of the TCP header before making the other checks.
			 */
			ip_len = ntohs(ipHeader->ip_len);
			if ((ip_len > protolen) ||
			    ((ip_len - (ipHeader->ip_hl << 2)) < 14)) {
				sc->stats.f_insideShortTcpHdr++;
				syslogMessage(SYSL_OUT_LENGTH, IPPROTO_TCP,
					srcAddr, dstAddr,
					ntohs(tcpHeader->th_sport),
					ntohs(tcpHeader->th_dport));				
				return DISCARD_PKT;
			}

			/* Check for "ACKless SYN" */
			if ((tcpHeader->th_flags & (TH_SYN|TH_ACK)) == TH_SYN) {
				return checkOutgoingTcp(srcAddr, dstAddr,
						ntohs(tcpHeader->th_sport),
						ntohs(tcpHeader->th_dport), sc);
			}
			break;

		case IPPROTO_UDP:
			udpHeader = (struct udphdr *)(pkt + 
						(ipHeader->ip_hl << 2));
			/*
			 * Make sure this packet (fragment) includes enough
			 * of the UDP header before making the other checks.
			 */
			ip_len = ntohs(ipHeader->ip_len);
			if ((ip_len > protolen) ||
			    ((ip_len - (ipHeader->ip_hl << 2)) < 4)) {
				sc->stats.f_insideShortUdpHdr++;
				syslogMessage(SYSL_OUT_LENGTH, IPPROTO_UDP,
					srcAddr, dstAddr,
					ntohs(udpHeader->uh_sport),
					ntohs(udpHeader->uh_dport));			
				return DISCARD_PKT;
			}

			return  checkOutgoingUdp( srcAddr, dstAddr,
					ntohs(udpHeader->uh_sport),
					ntohs(udpHeader->uh_dport), sc);
			break;

		case IPPROTO_ICMP:
			icmpHeader = (struct icmp *)
				     (pkt + (ipHeader->ip_hl << 2));

			/*
			 * If option is set, then discard fragmented ICMP pkts
			 */
			if (sc->flags.discardFragmentedICMP &&
			    (ipHeader->ip_off & IP_MF) != 0) {
				sc->stats.f_insideIcmpFrag++;	
				syslogMessage(SYSL_OUT_FRAG, IPPROTO_ICMP,
					srcAddr, dstAddr);		    	
				return DISCARD_PKT;
			}

			/*
			 * If option is set, discard attack ICMP pkts.
			 *
			 * If the ICMP type, code, identifier, and seq num all
			 * equal 0, this may be an ICMP echo reply attack pkt.
			 */
			if (sc->flags.discardAttackICMP &&
			    (*((u_int16_t *)icmpHeader) == 0) && /*type & code*/
			    (icmpHeader->icmp_void == 0)) {      /*id & seq*/

				if (icmpHeader->icmp_cksum == 0xFFFF) {
				
					/* This is a "smurf" attack */
					sc->stats.f_insideSmurfDos++;
					syslogMessage(SYSL_OUT_DOS,
						      srcAddr, dstAddr,
						      DOS_SMURF);			    		
					return DISCARD_PKT;

				} else if (*((u_int16_t *)icmpHeader->icmp_data)
					   == htons(0x4500)) {
					/* This is a "pong" attack */
					sc->stats.f_insidePongDos++;	
					syslogMessage(SYSL_OUT_DOS,
						      srcAddr, dstAddr,
						      DOS_PONG);		    		
					return DISCARD_PKT;
				}
			}

			return checkOutgoingIcmp(srcAddr, dstAddr,
						 icmpHeader->icmp_type, sc);
			break;

		default:
			if (sc->flags.discardOtherIp) {

				sc->stats.f_insideOtherIp++;	
				syslogMessage(SYSL_OUT_PROT,(int)ipHeader->ip_p,
					srcAddr, dstAddr);			
				return DISCARD_PKT;
			}

			/* Everything is allowed so far. */
			break;
		}

		break;

	case ETHERTYPE_ARP:
	case ETHERTYPE_REVARP:
		/* pass these guys through no matter what */
		break;

	default:
		/* Only pass the rest if told to do so. */
		if (sc->flags.discardNonIp) {

			sc->stats.f_insideNonIp++;
			syslogMessage(SYSL_OUT_FILTER, (int)protocol);		
			return DISCARD_PKT;
		}
	}

	return ALLOW_PKT;
}

/*
 * Main bridge function - called for packets not originated by drawbridge.
 * Parameters 'src' and 'dst' are expected to be in network byte order
 */

static int
bridge(int interface, HardwareAddr *src, HardwareAddr *dst, const sc_p sc)
{
	int pass = YES;

	/*
	 * Check if the destination address is a broadcast or
	 * multicast address and always forward if so.
	 */
	if (!IS_BROADCAST(dst)) {
		register int curr;
		register int hash;

		/*
		 * Check the bridge table for the destination address.
		 */
		hash = (dst->words[0] ^ dst->words[1] ^ dst->words[2])
			& ADDRESS_HASH_MASK;

		for (curr = hash; hashTable[curr].valid; ) {

			if (IS_HW_ADDR_EQU(dst, hashTable[curr].address)) {
			    	/*
				 * We found the address and know which side of
				 * the bridge it lives on.  Only forward if
				 * this packet is for a host on the other side.
				 */
				if (interface == hashTable[curr].interface)
					pass = NO;
				break;
			}

			/*
			 * Increment 'curr' and check to see if we have wrapped.
			 * If so, the table is full and the address was not
			 * found so forward the packet regardless.
			 */
			if ((curr = (curr+1) & ADDRESS_HASH_MASK) == hash)
				break;
		}
	}

	/*
	 * Now check the source address and make sure it's in the bridge
	 * table.  Checking for broadcast is somewhat superfluous but if
	 * some broken implementation sources a broadcast or multicast
	 * address, we sure don't want to add it to the bridge table!
	 */
	if (!IS_BROADCAST(src)) {
		register int curr;
		register int hash;

		/*
		 * Check the bridge table for the source address.
		 */
		hash = (src->words[0] ^ src->words[1] ^ src->words[2])
			& ADDRESS_HASH_MASK;

		for (curr = hash; ; ) {

			if (IS_HW_ADDR_EQU(src, hashTable[curr].address)) {
				
			    	/*
				 * It's already in the table so just make
				 * sure that the side of the bridge it's on
				 * stays up to date.
				 */
				hashTable[curr].interface = interface;
				
			    	break;
			}

			/*
			 * If the above compare was false, and this slot is
			 * open, then the address is not in the table and
			 * we can just instert it here.
			 */
			if (hashTable[curr].valid == NO) {

                                hashTable[curr].address.addr = src->addr;
                                hashTable[curr].interface = interface;
                                hashTable[curr].valid = YES;

				sc->stats.bridgeEntries++;
				break;
			}

			/*
			 * Increment 'curr' and check to see if we have wrapped.
			 * If so, then the table is already full.
			 */
			if ((curr = (curr+1) & ADDRESS_HASH_MASK) == hash)
				break;
		}
	}

	return pass;
}

int
drawbridge_test_pkt(const sc_p sc, struct hookinfo *const hinfo,struct mbuf *m)
{
	u_int16_t protocol;
	u_int8_t *pkt;
	u_int protolen, pktlen;
	HardwareAddr *dhost;
	HardwareAddr *shost;
	struct ether_header *h;
	register int rc;
	
	/* Get packet header. */
	h = mtod(m,struct ether_header *);

	/* Media type must be ethernet. */
	dhost = (HardwareAddr *)((struct ether_header *)h)->ether_dhost;
	shost = (HardwareAddr *)((struct ether_header *)h)->ether_shost;
	protocol = ntohs(((struct ether_header *)h)->ether_type);
	pkt = m->m_data+sizeof(struct ether_header); 
	protolen = m->m_pkthdr.len-sizeof(struct ether_header);
	pktlen = m->m_pkthdr.len;		
	
	/* Process packet from the inside interface. */
	if (hinfo == &sc->inside) { 
		
		/* Update stats. */
		sc->stats.p_insideRx++;
		sc->stats.b_insideRx += pktlen;	     
		
		/*
		 * Check the bridge table to see if the destination host is
		 * on the same side as the source host.  If so, drop the pkt.
		 */
		if (bridge(DB_INSIDE_IF,shost,dhost,sc) == NO)
			return DB_DROP_PKT;
		
		/* Check to see if the packet should be allowed through.*/		
		rc = checkOutgoingPacket(protocol,pkt,protolen,sc);
		if (rc != ALLOW_PKT) {
			sc->stats.p_insideFiltered++;
			sc->stats.b_insideFiltered += pktlen;
			if (rc == DISCARD_AND_ANSWER_PKT)
				return DB_DROP_AND_ANSWER_PKT;
			return DB_DROP_PKT;
		}

		/* Update stats and return. */
		sc->stats.p_outsideTx++;
		sc->stats.b_outsideTx += pktlen;		
		return DB_FORWARD_PKT;

	} else if (hinfo == &sc->outside) {  /* from the outside interface */	

		/* Update stats. */
		sc->stats.p_outsideRx++;
		sc->stats.b_outsideRx += pktlen;
		
		/*
		 * Check the bridge table to see if the destination host is
		 * on the same side as the source host.  If so, drop the pkt.
		 */
		if (bridge(DB_OUTSIDE_IF,shost,dhost,sc) == NO)
			return DB_DROP_PKT;

		/* Check to see if the packet should be allowed through. */
		rc = checkIncomingPacket(protocol,pkt,protolen,0,sc);
		if (rc != ALLOW_PKT) {
			sc->stats.p_outsideFiltered++;
			sc->stats.b_outsideFiltered += pktlen;
			if (rc == DISCARD_AND_ANSWER_PKT)
				return DB_DROP_AND_ANSWER_PKT;
			return DB_DROP_PKT;			
		}

		/* Update stats and return. */
		sc->stats.p_insideTx++;
		sc->stats.b_insideTx += pktlen;	     
		return DB_FORWARD_PKT;
	}
	return(0);
}

void
drawbridge_response_pkt(struct mbuf *r, struct mbuf *m)
{
	u_int32_t tmp;
	struct pkt_struct {
		struct ip i;
		struct tcphdr t;
	};
	register struct pkt_struct *out_pkt;
	register struct pkt_struct *in_pkt;
	register struct ether_header *in_h = mtod(m, struct ether_header *);
	register struct ether_header *out_h = mtod(r, struct ether_header *);
		
	memcpy(out_h->ether_dhost, in_h->ether_shost, 6);
	memcpy(out_h->ether_shost, in_h->ether_dhost, 6);
	out_h->ether_type = in_h->ether_type;

	r->m_data += sizeof(struct ether_header);
	m->m_data += sizeof(struct ether_header);
	r->m_pkthdr.rcvif = NULL;

	/* Setup pointers and clear outbound packet */
	in_pkt = mtod(m, struct pkt_struct *);
	out_pkt = mtod(r, struct pkt_struct *);
	r->m_len = sizeof(struct pkt_struct);
	bzero(out_pkt, r->m_len);

	/* IP header fields included in the TCP checksum */
	out_pkt->i.ip_len = htons(20);		/* TCP len */
	out_pkt->i.ip_p = IPPROTO_TCP;
	out_pkt->i.ip_src = in_pkt->i.ip_dst;
	out_pkt->i.ip_dst = in_pkt->i.ip_src;

	/* TCP header */
	out_pkt->t.th_sport = in_pkt->t.th_dport;
	out_pkt->t.th_dport = in_pkt->t.th_sport;
	out_pkt->t.th_seq = in_pkt->t.th_ack;
	tmp = ntohl(in_pkt->t.th_seq) + 1;
	out_pkt->t.th_ack = htonl(tmp);
	out_pkt->t.th_off = 5;
	out_pkt->t.th_flags = 0x14;

	/* TCP checksum */
	out_pkt->t.th_sum = in_cksum(r, r->m_len);

	/* Finish the IP header */
	out_pkt->i.ip_v = 4;
	out_pkt->i.ip_hl = 5;			/* IP hdr len >> 2 */
	out_pkt->i.ip_len = htons(40);		/* Total length */
	out_pkt->i.ip_ttl = 128;

	/* IP header checksum */
	out_pkt->i.ip_sum = in_cksum_hdr(&out_pkt->i);

	/* Adjust outbound mbuf to include media header */	
	r->m_data -= sizeof(struct ether_header);
	r->m_len += sizeof(struct ether_header);
	r->m_pkthdr.len = r->m_len;
}

static void
syslogMessage(u_int eventNo,...)
{
        va_list ap;
	int protocolNo;
	in_addr_tw srcAddr;
	in_addr_tw dstAddr;
	int srcPort;
	int dstPort;
	int type;
	int protocol;

	/*
	 * If this message is disabled then just return.
	 */
	if ( !((1UL << eventNo) & logMask) )
		return;

	switch ( eventNo ) {
		case SYSL_IN_OFFSET:
		case SYSL_OUT_OFFSET:
		case SYSL_IN_FRAG:
		case SYSL_OUT_FRAG:
		case SYSL_IN_REJECT:
		case SYSL_OUT_ACCEPT:
		case SYSL_IN_PROT:
		case SYSL_OUT_PROT: {	/* IP - general */
			va_start(ap,eventNo);

			protocolNo = va_arg(ap,int);
			srcAddr    = va_arg(ap,in_addr_tw);
			dstAddr    = va_arg(ap,in_addr_tw);

			va_end(ap);

			log(	(logFacility |
				 syslogMessages[eventNo].priority),
				"drawbridge: %s protocol %d " \
				"from %u.%u.%u.%u to %u.%u.%u.%u\n",
				syslogMessages[eventNo].message, protocolNo,
				srcAddr.s_b[3], srcAddr.s_b[2],
				srcAddr.s_b[1], srcAddr.s_b[0],
				dstAddr.s_b[3], dstAddr.s_b[2],
				dstAddr.s_b[1], dstAddr.s_b[0]
			);
			break;
		}

		case SYSL_IN_CLASSD:
		case SYSL_OUT_CLASSD:
		case SYSL_IN_PORT:
		case SYSL_OUT_PORT:
		case SYSL_OUT_OVERRIDE:
		case SYSL_IN_LENGTH:
		case SYSL_OUT_LENGTH: {	/* IP - TCP/UDP */
			va_start(ap,eventNo);

			protocolNo = va_arg(ap,int);
			srcAddr    = va_arg(ap,in_addr_tw);
			dstAddr    = va_arg(ap,in_addr_tw);
			srcPort    = va_arg(ap,int);
			dstPort    = va_arg(ap,int);

			va_end(ap);

			log(	(logFacility |
				 syslogMessages[eventNo].priority),
				"drawbridge: %s %s from %u.%u.%u.%u " \
				"port %u to %u.%u.%u.%u port %u\n",
				(protocolNo == IPPROTO_UDP) ? "UDP" : "TCP",
				syslogMessages[eventNo].message,
				srcAddr.s_b[3], srcAddr.s_b[2],
				srcAddr.s_b[1], srcAddr.s_b[0],
				srcPort,
				dstAddr.s_b[3], dstAddr.s_b[2],
				dstAddr.s_b[1], dstAddr.s_b[0],
				dstPort
			);
			break;
		}

		case SYSL_IN_TYPE: 	/* IP - ICMP */
		case SYSL_OUT_TYPE: {
			va_start(ap,eventNo);

			srcAddr = va_arg(ap,in_addr_tw);
			dstAddr = va_arg(ap,in_addr_tw);
			type    = va_arg(ap,int);

			va_end(ap);

			log(	(logFacility |
				 syslogMessages[eventNo].priority),
				"drawbridge: ICMP %s from " \
				"%u.%u.%u.%u type %u to %u.%u.%u.%u\n",
				syslogMessages[eventNo].message,
				srcAddr.s_b[3], srcAddr.s_b[2],
				srcAddr.s_b[1], srcAddr.s_b[0],
				type,
				dstAddr.s_b[3], dstAddr.s_b[2],
				dstAddr.s_b[1], dstAddr.s_b[0]
			);
			break;
		}

		case SYSL_IN_DOS:	/* Denial of service attack */
		case SYSL_OUT_DOS: {
			va_start(ap,eventNo);

			srcAddr = va_arg(ap,in_addr_tw);
			dstAddr = va_arg(ap,in_addr_tw);
			type    = va_arg(ap,int);

			va_end(ap);

			log(	(logFacility |
				syslogMessages[eventNo].priority),
				"drawbridge: %s %s from " \
				"%u.%u.%u.%u to %u.%u.%u.%u\n",
				syslogDosTypes[type],
				syslogMessages[eventNo].message,
				srcAddr.s_b[3], srcAddr.s_b[2],
				srcAddr.s_b[1], srcAddr.s_b[0],
				dstAddr.s_b[3], dstAddr.s_b[2],
				dstAddr.s_b[1], dstAddr.s_b[0]
			);
			break;
		}

		case SYSL_IN_FILTER:
		case SYSL_OUT_FILTER: {	/* MAC layer protocol */
			va_start(ap,eventNo);
			protocol = va_arg(ap,int);
			va_end(ap);

			log(	(logFacility |
				 syslogMessages[eventNo].priority),
				"drawbridge: %s 0x%04x\n",
				syslogMessages[eventNo].message, protocol
			);
			break;
		}

		case SYSL_COLD_START:	/* Startup message */
			log(	(logFacility |
				 syslogMessages[eventNo].priority),
				"drawbridge: %s\n",
				syslogMessages[eventNo].message
			);
			break;
			
		default:		/* Should never get here */
			log(	(logFacility | LOG_WARNING),
				"drawbridge: unknown event\n"
			);
			break;
	}
}








